import numpy as np
from . import cosutil

class Dispersion(object):
    """Dispersion relation.

    The public methods are:
        flag = disprel.isValid()
        nrows = disprel.getNRows()
        filter = disprel.getFilter()
        wavelength = disprel.evalDisp(x)
        dwavelength / dx = disprel.evalDerivDisp(x)
        x = disprel.evalInvDisp(wavelength, tiny=1.e-8)
        disprel.info()
        disprel.close()

    Parameters
    ----------
    disptab: str
        name of table containing dispersion relations

    filter: dictionary
        parameters for selecting a row from the disptab

    use_fpoffset: boolean
        if True, include fpoffset in the filter;
        if False, exclude it from the filter
    """

    def __init__(self, disptab, filter, use_fpoffset=True):

        # This will be a local copy of filter, possibly excluding "fpoffset".
        self.filter = {}
        self.use_fpoffset = use_fpoffset
        self.ncoeff = 0
        self.coeff = None
        self.delta = 0.
        self.fpoffset = 0               # save for information
        self._nrows = 0                 # number of matching rows (should be 1)
        self._valid = True

        for key in filter.keys():
            key_lower = key.lower()
            if key_lower == "fpoffset":
                self.fpoffset = filter[key]
                continue
            self.filter[key_lower] = filter[key]
        if use_fpoffset:
            self.filter["fpoffset"] = self.fpoffset

        if not cosutil.findColumn(disptab, "fpoffset"):
            if "fpoffset" in self.filter:
                del(self.filter["fpoffset"])

        disp_info = cosutil.getTable(disptab, self.filter)
        if disp_info is None:
            self._valid = False
            del disp_info
            return
        else:
            self._nrows = len(disp_info)

        self.ncoeff = disp_info.field("nelem")[0]
        if self.ncoeff < 2:
            raise ValueError("Dispersion relation has too few coefficients")
        self.coeff = disp_info.field("coeff")[0][0:self.ncoeff]
        if cosutil.findColumn(disp_info, "delta"):
            self.delta = disp_info.field("delta")[0]
        else:
            if cosutil.findColumn(disp_info, "d_tv03"):
                d_tv03 = disp_info.field("d_tv03")[0]
            else:
                d_tv03 = 0.
            if cosutil.findColumn(disp_info, "d"):
                d = disp_info.field("d")[0]
            else:
                d = 0.
            self.delta = d_tv03 - d

        del disp_info

    def info(self):

        cosutil.printMsg("filter = %s" % str(self.filter))
        cosutil.printMsg("use_fpoffset = %s" % str(self.use_fpoffset))
        cosutil.printMsg("fpoffset = %d" % self.fpoffset)
        cosutil.printMsg("number of coefficients = %d" % self.ncoeff)
        cosutil.printMsg("coeff = %s" % str(self.coeff))
        cosutil.printMsg("delta = %.6g" % self.delta)
        cosutil.printMsg("number of matching rows = %d" % self._nrows)
        cosutil.printMsg("valid = %s" % str(self._valid))

    def isValid(self):
        """Return True if a matching row was found in disptab."""

        return self._valid

    def getNRows(self):
        """Return the number of rows in disptab that match the filter."""

        return self._nrows

    def getFilter(self):
        """Return the filter dictionary."""

        return self.filter

    def close(self):
        """Delete coefficients and reset attributes."""

        del self.coeff
        self.filter = {}
        self.ncoeff = 0
        self.delta = 0.
        self.fpoffset = 0
        self._nrows = 0
        self._valid = False

    def evalDisp(self, x):
        """Evaluate the dispersion relation at x.

        The function value will be the wavelength (or array of wavelengths)
        at x, in Angstroms.

        Parameters
        ----------
        x: array_like or float
            Pixel coordinate (or array of coordinates)

        Returns
        -------
        array_like or float
            Wavelength (or array of wavelengths) at x
        """

        x_prime = np.float64(x) + np.float64(self.delta)

        sum = self.coeff[self.ncoeff-1]
        for i in range(self.ncoeff-2, -1, -1):
            sum = sum * x_prime + self.coeff[i]

        return sum

    def evalDerivDisp(self, x):
        """Evaluate the derivative of the dispersion relation at x.

        The function value will be the slope (or array of slopes) at x,
        in Angstroms per pixel.

        Parameters
        ----------
        x: array_like or float
            Pixel coordinate (or array of coordinates)

        Returns
        -------
        array_like or float
            Slope at x, in Angstroms per pixel
        """

        x_prime = np.float64(x) + np.float64(self.delta)

        sum = (self.ncoeff - 1.) * self.coeff[self.ncoeff-1]
        for n in range(self.ncoeff-2, 0, -1):
            sum = sum * x_prime + n * self.coeff[n]

        return sum

    def evalInvDisp(self, wavelength, tiny=1.e-8):
        """Evaluate the inverse of the dispersion relation at wavelength.

        The function value will be the pixel number (or array of pixel numbers)
        at the specified wavelength(s).  Newton's method is used for finding
        the pixel numbers, and the iterations are stopped when the largest
        difference between the specified wavelengths and computed wavelengths
        is less than tiny.

        Parameters
        ----------
        wavelength: array_like or float
            Wavelength or array of wavelengths

        tiny: float
            Maximum allowed difference between the final pixel number(s)
            and the value from the previous iteration

        Returns
        -------
        array_like or float
            Pixel number (or array of pixel numbers) at wavelength
        """

        tiny = abs(tiny)

        # initial value
        try:
            nelem = len(wavelength)
            x = np.arange(nelem, dtype=np.float64)
        except TypeError:
            nelem = 0
            x = 0.

        # Iterate to find the pixel number(s) x such that evaluating the
        # dispersion relation at that point or points gives the specified
        # wavelength(s).
        done = False
        while not done:
            if nelem > 0:
                x_prev = x.copy()
            else:
                x_prev = x
            wl = self.evalDisp(x)
            slope = self.evalDerivDisp(x)
            wl_diff = wavelength - wl
            x += wl_diff / slope
            diff = np.abs(x - x_prev)
            if diff.max() < tiny:
                done = True

        return x



import numpy as np
from scipy.signal.windows import boxcar
from scipy import ndimage
from . import cosutil
from .calcosparam import *       # parameter definitions
from . import ccos

MASK_X = 189            # width of region to mask for each geocoronal line
SEARCH_Y = 91           # height of search region

# for comparison between values of fwhm
SIGNFICANTLY_LARGER = 2.        # pixels

# Lyman alpha, oxygen I, oxygen I
AIRGLOW_WAVELENGTHS = [1215.67, 1304., 1356.]

"""Info pertaining to geocoronal features:
For G130M, 1309, PSA, the image of geocoronal LyA is about 184 x 60 pixels.
The PSA and WCA apertures are about 100 pixels apart, for all FUV gratings.
NUVA and NUVB are closer, about 94 pixels for the M gratings.

The plate scale (arcsec per pixel) in the dispersion direction is roughly
the same for all gratings; it should be OK to mask a region 189 pixels wide,
centered on each geocoronal line.

A search range of 91 pixels seems reasonable (-45 to +45 pixels inclusive).

In t9h1220sl_phot.fits for NUV, the minimum wavelength for the M gratings
is 1648.58 Angstroms.  Wavelengths go down to 1213.77 for G230L, 2635,
but the throughput at short wavelengths is very low.
"""

def xdSearch(data, dq_data, wavelength, axis, slope, y_nominal,
             x_offset, detector):
    """Find the cross-dispersion location of the target spectrum.

    Parameters
    ----------
    data: array_like, 2-D
        SCI data from the flt file.

    dq_data: array_like, 2-D
        DQ data from the flt file.

    wavelength: array_like, 1-D
        Wavelength at each pixel (only needed for FUV).

    axis: int
        The dispersion axis, 0 (Y) or 1 (X).

    slope: float
        Slope of spectrum, pixels per pixel.

    y_nominal: float
        Location of spectrum at left edge of detector, i.e. at
        X = x_offset.

    x_offset: intls
        Offset of the detector in the data array.

    detector: str
        Detector name ("FUV" or "NUV").

    Returns
    -------
    (offset2, y_locn, y_locn_sigma, fwhm): tuple of four values
        offset2 is the offset of the spectrum from y_nominal (positive
        if the spectrum was found at a larger Y pixel number).  y_locn
        is the Y pixel number at which the spectrum was found (at pixel
        x_offset from the left edge of data).  y_locn_sigma is the
        error estimate for y_locn.  fwhm is the full-width at half
        maximum of the peak in the cross-dispersion profile; this can be
        either a float or an int.
    """

    (e_j, zero_point) = extractBand(data, dq_data, wavelength,
                                    axis, slope, y_nominal,
                                    x_offset, detector)

    box = 3

    (y_locn, y_locn_sigma, fwhm) = findPeak(e_j, box)

    if y_locn is None:
        offset2 = 0.
    else:
        # Shift y_locn to account for the offset of e_j from Y = 0 in 'data',
        # and shift y_locn to where the spectrum crosses X = x_offset.
        y_locn += zero_point
        y_locn += slope * float(x_offset)
        offset2 = y_locn - y_nominal

    return (offset2, y_locn, y_locn_sigma, fwhm)

def extractBand(data, dq_data, wavelength, axis, slope, y_nominal,
                x_offset, detector):
    """Extract a 2-D stripe centered on the nominal location of the target.

    Parameters
    ----------
    data: array_like
        SCI data from the flt file

    dq_data: array_like
        DQ data from the flt file

    wavelength: array_like
        Wavelength at each pixel (to locate the airglow lines)

    axis: int
        The dispersion axis, 0 (Y) or 1 (X)

    slope: float
        Slope of spectrum, pixels per pixel

    y_nominal: float
        Intercept of spectrum at left edge of detector

    x_offset: int
        Offset of the detector in the data array

    detector: str
        Detector name ("FUV" or "NUV")

    Returns
    -------
    tuple
        (e_j, zero_point), where e_j is a 1-D array containing a section of
        data collapsed along the dispersion direction and zero_point is
        the Y pixel number at the left edge of data corresponding to
        pixel 0 of e_j
    """

    extr_height = SEARCH_Y
    axis_length = data.shape[axis]
    e_ij = np.zeros((extr_height, axis_length), dtype=np.float32)
    ccos.extractband(data, axis, slope, y_nominal, x_offset, e_ij)

    # Clobber any region flagged as bad; note that this won't work well if a
    # flagged region covers part but not all of a spectral feature.
    if dq_data is not None:
        dq_ij = np.zeros((extr_height, axis_length), dtype=np.int16)
        ccos.extractband(dq_data, axis, slope, y_nominal, x_offset, dq_ij)
        dq = np.where(dq_ij == 0, 1, 0)
        e_ij *= dq

    if detector == "FUV":
        # Block out (i.e. set to zero) regions affected by airglow lines.
        for airglow in AIRGLOW_WAVELENGTHS:
            pixel_center = findPixelNumber(wavelength, airglow)
            pixel0 = pixel_center - (MASK_X // 2)
            pixel1 = pixel_center + (MASK_X // 2)
            if pixel1 < 0 or pixel0 >= axis_length:
                continue
            pixel0 = max(pixel0, 0)
            pixel1 = min(pixel1, axis_length-1)
            e_ij[:,int(pixel0):int(pixel1)] = 0.

    # sum the data along the dispersion direction
    e_j = e_ij.sum(axis=1)

    # Y pixel number in data corresponding to e_j[0]
    zero_point = int(round(y_nominal - slope * float(x_offset))) - \
                 SEARCH_Y // 2

    return (e_j, zero_point)

def findPixelNumber(wl, wavelength):
    """Find the nearest pixel to 'wavelength'.

    Parameters
    ----------
    wl: array_like, float64
        Wavelength at each pixel, assumed to be increasing

    wavelength: float
        A particular wavelength

    Returns
    -------
    int
        Pixel number closest to wavelength in the array wl
    """

    nelem = len(wl)

    dispersion = (wl[-1] - wl[0]) / float(nelem)
    if wavelength < wl[0]:
        x = (wavelength - wl[0]) / dispersion
        return int(round(x))
    elif wavelength >= wl[-1]:
        x = (wavelength - wl[-1]) / dispersion + float(nelem) - 1.
        return int(round(x))

    i0 = 0
    i1 = nelem - 1
    while (i1 - i0) >= 5:
        if i0 == i1:
            break
        slope = (wl[i1] - wl[i0]) / (i1 - i0)
        if slope == 0.:
            raise RuntimeError("Bad wavelength array.")
        mid = (i1 + i0) // 2
        x = int(round((wavelength - wl[mid]) / slope)) + mid
        dx = i1 - i0
        i0 = x - dx // 16
        i1 = x + dx // 16
        i0 = max(i0, 0)
        i1 = min(i1, nelem-1)

    x = i0
    diff = abs(wavelength - wl[x])
    for i in range(i0, i1+1):
        if abs(wavelength - wl[i]) < diff:
            x = i
            diff = abs(wavelength - wl[x])

    return x

def findPeak(e_j, box):
    """Find the location of the maximum within the subset.

    Note that the data were collapsed to the left edge to get e_j, so
    the location is the intercept on the edge, rather than where the
    spectrum crosses the middle of the detector or where it crosses
    X = x_offset.
    Also, e_j is not the full height of the detector, just a subset
    centered on the nominal Y location of the spectrum.

    Parameters
    ----------
    e_j: array_like
        1-D array of data collapsed along dispersion axis, taking into
        account the tilt of the spectrum

    box: int
        Smooth e_j with a box of this width before looking for the
        maximum

    Returns
    -------
    tuple
        The location (float) in the cross-dispersion direction relative
        to the first pixel in e_j, an estimate of the uncertainty in
        that location, and the FWHM of the peak in the cross-dispersion
        profile
    """

    boxcar_kernel = boxcar(box) / box
    e_j_sm = ndimage.convolve(e_j, boxcar_kernel, mode="nearest")

    index = np.argsort(e_j_sm)
    ymax = index[-1]

    nelem = len(e_j)

    # This may be done again later, after we have found the location more
    # accurately.
    fwhm = findFwhm(e_j, ymax)

    # fit a quadratic to at least five points centered on ymax
    MIN_NPTS = 5
    npts = int(round(fwhm))
    npts = max(npts, MIN_NPTS)
    if npts // 2 * 2 == npts:
        npts += 1
    x = np.arange(nelem, dtype=np.float64)
    j1 = ymax - npts // 2
    j1 = max(j1, 0)
    j2 = j1 + npts
    if j2 > nelem:
        j2 = nelem
        j1 = j2 - npts
        j1 = max(j1, 0)
    (coeff, var) = cosutil.fitQuadratic(x[j1:j2], e_j_sm[j1:j2])

    (y_locn, y_locn_sigma) = cosutil.centerOfQuadratic(coeff, var)
    if y_locn is None:
        y_locn = ymax
        y_locn_sigma = 999.

    # Find the FWHM again if the location is far from the brightest pixel.
    if abs(y_locn - ymax) > fwhm / 4.:
        fwhm = findFwhm(e_j, y_locn)

    return (y_locn, y_locn_sigma, fwhm)

def findFwhm(e_j, y_locn):
    """Find the FWHM of the cross-dispersion profile of the spectrum.

    Two different approaches are used to find the FWHM.  The first method
    is to count the number of elements in the cross-dispersion profile with
    values above the half-maximum value; this value will be an integer.
    The second method is to follow the profile to the half-maximum value on
    either side of the maximum, using linear interpolation to get a better
    estimate of where the profile cuts across the half-maximum level; this
    value will be a float.  The value from the second method is expected to
    be more accurate if the target was actually found and has good
    signal-to-noise, so normally that value will be returned.  If the first
    method gives a significantly larger value, however, that value will be
    returned because it may indicate that the cross-dispersion profile is
    just noise.

    Parameters
    ----------
    e_j: array_like
        1-D array of data collapsed along dispersion axis

    y_locn: float
        The location in the cross-dispersion direction, relative to the
        first pixel in e_j.

    Returns
    -------
    float or int
        The full width half maximum of the peak in e_j.
    """

    nelem = len(e_j)
    y_locn_nint = int(round(y_locn))
    if y_locn_nint < 0 or y_locn_nint >= nelem:
        return -1.

    e_max = e_j[y_locn_nint]
    if e_max <= 0:
        return -1.

    e_j_sorted = np.sort(e_j)

    third = nelem // 3
    background = e_j_sorted[0:third].mean(dtype=np.float64)

    find_this_level = (e_max - background) / 2. + background

    # first estimate of FWHM
    # Count all elements in the sorted array that are greater than the
    # halfway level.  This will be large if the array is just noise
    # (at least, that's the idea).
    j = nelem - 1
    while j >= 0:
        if e_j_sorted[j] < find_this_level:
            break
        j -= 1
    fwhm_1 = nelem - 1 - j      # this is an int

    # second estimate of FWHM
    # Find where the cross-dispersion profile crosses the halfway level
    # on either side of the maximum.

    j_low = 0                   # initial values
    j_high = nelem - 1

    # first the low side
    j = y_locn_nint
    while j >= 0:
        if e_j[j] < find_this_level:
            j_low = j
            break
        j -= 1

    # Use linear interpolation to find where e_j would equal find_this_level.
    denom = e_j[j_low+1] - e_j[j_low]
    if denom == 0.:
        low = float(j_low) + 0.5        # 0.5 is an estimate
    else:
        low = float(j_low) + (find_this_level - e_j[j_low]) / denom

    # now the high side
    j = y_locn_nint
    while j < nelem:
        if e_j[j] < find_this_level:
            j_high = j
            break
        j += 1

    denom = e_j[j_high] - e_j[j_high-1]
    if denom == 0.:
        high = float(j_high) - 0.5
    else:
        high = float(j_high-1) + (find_this_level - e_j[j_high-1]) / denom

    fwhm_2 = high - low         # this is a float

    if fwhm_1 > fwhm_2 + SIGNFICANTLY_LARGER:
        fwhm = fwhm_1
    else:
        fwhm = fwhm_2

    return fwhm



# This file defines parameters used by calcos.
from . import __version__

CALCOS_VERSION_NUMBER = __version__
CALCOS_VERSION = "%s" % (CALCOS_VERSION_NUMBER)

# These are the values to indicate the detector (original) and user
# (flipped or rotated) COS coordinates.
DETECTOR_COORDINATES = "DETECTOR"
USER_COORDINATES     = "USER"

SPEED_OF_LIGHT = 299792.458     # km/s

DAYS_PER_YEAR = 365.25
SEC_PER_DAY = 86400.

MJD_TO_JD = 2400000.5           # add to MJD to get Julian Day Number

# Live time estimates should not differ by more than this fraction of
# the live time.
LIVETIME_CRITERION = 0.1

# This is the wavelength below which no significant flux could be detected.
MIN_WAVELENGTH = 900.           # Angstroms

# These give the axis lengths of the FUV and NUV detectors, in pixels.
FUV_X = 16384                   # more rapidly varying axis
FUV_Y = 1024
NUV_X = 1024                    # more rapidly varying axis
NUV_Y = 1024

# X_OFFSET is the offset of the detector in a calibrated image.
# Pixel X in a calibrated image = XFULL + X_OFFSET
FUV_X_OFFSET = 0
FUV_EXTENDED_X = FUV_X
NUV_X_OFFSET = 100
NUV_EXTENDED_X = NUV_X + 250

# These are the default binning factors for FUV and NUV "calcos sum" images.
FUV_BIN_X = 1
FUV_BIN_Y = 1
NUV_BIN_X = 1
NUV_BIN_Y = 1

# These give the number of spectra per detector (used in extract.py).
FUV_SPECTRA = 1                 # one spectrum on one FUV segment
NUV_SPECTRA = 3                 # three stripes on NUV detector

# These are the possible values for verbosity.
QUIET = 0
VERBOSE = 1
VERY_VERBOSE = 2

# These are the possible values for the TAGFLASH keyword, and corresponding
# integer codes.
TAGFLASH_NONE = "NONE"
TAGFLASH_AUTO = "AUTO"
TAGFLASH_UNIFORMLY_SPACED = "UNIFORMLY SPACED"
TAGFLASH_TYPE_NONE = 0
TAGFLASH_TYPE_AUTO = 1
TAGFLASH_TYPE_UNIFORMLY_SPACED = 2

# The following section pertains to changes in the position of the aperture
# block to move the target on the detector, to extend the lifetime of the
# FUV detector.

# These are the names of the apertures.
APERTURE_NAMES = ["PSA", "WCA", "BOA", "FCA"]
# The aperture keyword for a dark exposure may have this value instead.
OTHER_APERTURE_NAMES = ["N/A"]

# Nominal values of aperypos for life_adj = 1.
APERTURE_POSN1 = {"PSA": 126.,
                  "WCA": 126.,
                  "BOA": -153.,
                  "FCA": -153.}

# pixels per arcsecond in the cross-dispersion direction
XD_PLATE_SCALE = {
    "G130M": 11.9,      # email from Charles, 2012 Feb 2
    "G160M": 11.9,
    "G140L": 11.9,
    "G185M": 41.85,
    "G225M": 41.89,
    "G285M": 41.80,
    "G230L": 42.27,
    "MIRRORA": 42.5,    # COS ISR 2010-10
    "MIRRORB": 42.5}

# arcseconds per step of the aperture block in the cross-dispersion direction
ARCSEC_PER_XD_APER_STEP = -0.0476

# This value is read/write.  It can be set to a value other than zero
# if LIFE_ADJ = -1.
LIFE_ADJ_OFFSET = 0.            # pixels

# This is the list of segment-specific (or in some cases stripe-specific)
# keywords, with "X" (case sensitive) replaced by "a", "b" or "c".
segment_specific_keywords = \
    ["stimX_lx", "stimX_ly", "stimX_rx", "stimX_ry",
     "stimX0lx", "stimX0ly", "stimX0rx", "stimX0ry",
     "stimXslx", "stimXsly", "stimXsrx", "stimXsry",
     "npha_X", "phalowrX", "phaupprX",
     "tbrst_X", "nbrst_X", "tbadt_X", "nbadt_X",
     "nout_X", "nbadevtX",
     "exptimeX", "neventsX",
     "globrt_X",
     "deadrt_X", "deadmt_X", "livetm_X",
     "sp_loc_X", "sp_off_X", "sp_nom_X", "sp_slp_X", "sp_hgt_X", "sp_err_X",
     "b_bkg1_X", "b_bkg2_X",
     "b_hgt1_X", "b_hgt2_X",
     "shift1X", "shift2X", "dpixel1X",
     "chi_sq_X", "ndf_X"]

# The pulse height values range from 0 to 127.  The values in the PHA
# column of an EVENTS table, however, come from a five-bit value, i.e.
# the last two bits have been truncated, resulting in their being a
# factor of four smaller.
TWO_BITS = 4

# The following three parameters are used by getTable.
# NOT_APPLICABLE will be assigned as the value of a keyword that is
# missing from the header; this is done because some keywords may
# actually not be present, while others that are not relevant will be
# present but have the value "N/A".
STRING_WILDCARD = "ANY"
NOT_APPLICABLE = "N/A"
INT_WILDCARD = -1

# These are the data quality flags.
DQ_OK = 0                       # no anomalous condition noted
DQ_SOFTERR = 1                  # Reed-Solomon error
DQ_UNUSED_2 = 2                 # [currently unused]
DQ_DETECTOR_SHADOW = 4          # FUV grid shadow mark or NUV vignetting
DQ_POORLY_CALIBRATED = 8        # poorly calibrated (incl. detector edge)
DQ_VERY_LOW_RESPONSE = 16       # > 80% depression
DQ_BACKGROUND_FEATURE = 32      # background feature
DQ_BURST = 64                   # count rate implies a burst (FUV)
DQ_PIXEL_OUT_OF_BOUNDS = 128    # pixel out of bounds
DQ_DATA_FILL = 256              # fill data
DQ_PHA_OUT_OF_BOUNDS = 512      # pulse height is either too low or too high
DQ_LOW_RESPONSE_REGION = 1024   # > 50% depression
DQ_BAD_TIME = 2048              # time is within a bad time interval
DQ_LOW_PHA_FEATURE = 4096       # low PHA feature
DQ_GAIN_SAG_HOLE = 8192         # low gain area
DQ_UNUSED_16384 = 16384         # [currently unused]

# Use this when binning TIME-TAG data to images, or extracting spectra from
# TIME-TAG data.
SERIOUS_DQ_FLAGS = (DQ_BURST | DQ_BAD_TIME | DQ_PHA_OUT_OF_BOUNDS)

# Define an exception for the case that the APERTURE keyword is not recognized.
class BadApertureError(Exception):
    def __init__(self, message=None):
        self.message = message
    def __str__(self):
        return self.message

# Define an exception for a missing row in a reference file.
class MissingRowError(Exception):
    def __init__(self, message=None):
        self.message = message
    def __str__(self):
        return self.message

# Define an exception for a missing column in a reference file.
class MissingColumnError(Exception):
    def __init__(self, message=None):
        self.message = message
    def __str__(self):
        return self.message



"""Custom ``pytest`` configurations."""
import pytest
import os

#from astropy.tests.helper import enable_deprecations_as_exceptions

# Turn deprecation warnings into exceptions.
#enable_deprecations_as_exceptions()

# Require these pytest plugins to run.
pytest_plugins = ["pytest_ciwatson"]


# For easy inspection on what dependencies were used in test.
def pytest_report_header(config):
    import sys
    import warnings
    import importlib

    s = "\nFull Python Version: \n{0}\n\n".format(sys.version)

    for module_name in ('numpy', 'astropy', 'scipy', 'matplotlib',
                        'stsci.tools'):
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", DeprecationWarning)
                module = importlib.import_module(module_name)
        except ImportError:
            s += "{0}: not available\n".format(module_name)
        else:
            try:
                version = module.__version__
            except AttributeError:
                version = 'unknown (no __version__ attribute)'
            s += "{0}: {1}\n".format(module_name, version)

    return s


@pytest.fixture(scope='session')
def test_data():
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')



import os
from astropy.io import fits
from . import cosutil
from .calcosparam import *

def initialInfo(filename):
    """Get DETECTOR, OBSMODE, and EXPTYPE from the primary header.

    Parameters
    ----------
    filename: str
        Name of the input file.

    Returns
    -------
    info: dictionary
        Keywords and values (just a few) from the headers of the input
        file.
    """

    fd = fits.open(filename, mode="readonly")
    phdr = fd[0].header

    info = {}

    if "DETECTOR" in phdr:
        detector = phdr["DETECTOR"]
    else:
        raise RuntimeError("File " + filename +
                           " does not have DETECTOR keyword.")

    if "OBSMODE" in phdr:
        obsmode = phdr["OBSMODE"]
    else:
        raise RuntimeError("File " + filename +
                           " does not have OBSMODE keyword.")

    if "EXPTYPE" in phdr:
        exptype = phdr["EXPTYPE"]
    else:
        raise RuntimeError("File " + filename +
                           " does not have EXPTYPE keyword.")

    if detector != "FUV" and detector != "NUV":
        raise ValueError("File " + filename +
                         " has invalid DETECTOR = " + detector)

    if obsmode != "TIME-TAG" and obsmode != "ACCUM":
        raise ValueError("File " + filename +
                         " has invalid OBSMODE = " + obsmode)

    info["detector"] = detector
    info["obsmode"] = obsmode
    info["exptype"] = exptype

    fd.close()

    return info

def getGeneralInfo(phdr, hdr):
    """Get keyword values from the primary and extension header.

    The input argument phdr is the primary header, and the second hdr is
    the first extension header, as provided by the astropy fits module.  The
    function value is a dictionary of keyword = value pairs.  If a keyword
    is missing from the header, it will still be included in the dictionary,
    but its value will be set to the NOT_APPLICABLE string, or a reasonable
    default for keyword values that are not text strings.  npix (a tuple
    giving the output image size) will also be assigned.  If the data
    portion is empty (based on the NAXIS keyword), the value will be (0,);
    otherwise, the value will be assigned an appropriate value for the
    detector, rather than being read directly from the header.  The
    heliocentric velocity will be initialized to zero.

    Parameters
    ----------
    phdr: ``astropy.io.fits.header.Header`` object
        Primary header.

    hdr: ``astropy.io.fits.header.Header`` object
        Extension header.

    Returns
    -------
    info: dictionary
        Keywords and values from the headers of the input file.
    """

    info = {}

    # Get keywords from the primary header.

    # This is a list of primary header keywords and default values.
    keylist = {
        "detector":  NOT_APPLICABLE,
        "segment":   NOT_APPLICABLE,
        "obstype":   NOT_APPLICABLE,
        "obsmode":   NOT_APPLICABLE,
        "exptype":   NOT_APPLICABLE,
        "opt_elem":  NOT_APPLICABLE,
        "targname":  NOT_APPLICABLE,
        "lampused":  NOT_APPLICABLE,
        "lampplan":  NOT_APPLICABLE,
        "rootname":  "",
        "subarray":  False,
        "tagflash":  False,
        "cenwave":   0,
        "randseed": -1,
        "fppos":     1,
        "fpoffset":  0,
        "life_adj": -1,
        "aperypos":  NOT_APPLICABLE,
        "coscoord":  DETECTOR_COORDINATES,
        "ra_targ":  -999.,
        "dec_targ": -999.,
        "xtrctalg": "BOXCAR"}

    for key in keylist.keys():
        info[key] = phdr.get(key, default=keylist[key])

    # Set output image size (variables defined in calcosparam.py).
    if info["detector"] == "FUV":
        info["npix"] = (FUV_Y, FUV_EXTENDED_X)
        info["x_offset"] = FUV_X_OFFSET
    else:
        if info["obstype"] == "IMAGING":
            info["npix"] = (NUV_Y, NUV_X)
            info["x_offset"] = 0
        else:
            info["npix"] = (NUV_Y, NUV_EXTENDED_X)
            info["x_offset"] = NUV_X_OFFSET

    # Replace the value for npix if there's no data (based on extension header).
    if hdr["NAXIS"] == 0:
        info["npix"] = (0,)
    elif hdr["NAXIS"] == 2 and hdr["NAXIS2"] == 0:
        info["npix"] = (0,)

    # Assign an initial value for the heliocentric velocity
    info["v_helio"] = 0.

    (info["aperture"], message) = cosutil.getApertureKeyword(phdr)
    if message:
        cosutil.printWarning(message)

    # Update info["life_adj_offset"], if life_adj = -1.
    cosutil.computeLifeAdjOffset(info)

    if info["tagflash"] == TAGFLASH_AUTO:
        info["tagflash"] = True
        info["tagflash_type"] = TAGFLASH_TYPE_AUTO
    elif info["tagflash"] == TAGFLASH_UNIFORMLY_SPACED:
        info["tagflash"] = True
        info["tagflash_type"] = TAGFLASH_TYPE_UNIFORMLY_SPACED
    else:
        info["tagflash"] = False
        info["tagflash_type"] = TAGFLASH_TYPE_NONE

    # Engineering keywords relevant to deadtime correction.

    if info["detector"] == "FUV":
        if info["segment"] == "FUVA":
            countrate = phdr.get("DEVENTA", default=0.)
            info["countrate"] = hdr.get("DEVENTA", default=countrate)
        else:
            countrate = phdr.get("DEVENTB", default=0.)
            info["countrate"] = hdr.get("DEVENTB", default=countrate)
    else:
        countrate = phdr.get("MEVENTS", default=0.)
        info["countrate"] = hdr.get("MEVENTS", default=countrate)

    # Now get keywords from the extension header.

    if info["detector"] == "FUV":
        # The header keyword is the rate for both stims together; we want
        # the rate for one stim.
        info["stimrate"] = hdr.get("STIMRATE", default=0.) / 2.
    else:
        info["stimrate"] = 0.

    # This is a list of extension header keywords and default values.
    keylist = {
        "dispaxis":  0,
        "sdqflags":  184,       # 8 + 16 + 32 + 128
        "nsubarry":  0,
        "numflash":  0,
        "expstart": -1.,
        "expend":   -1.,
        "doppon":    False,
        "doppont":   False,
        "doppmagv": -1.,
        "dopmagt":  -1.,
        "doppzero": -1.,
        "dopzerot": -1.,
        "orbitper": -1.,
        "orbtpert": -1.,
        "ra_aper":   0.,
        "dec_aper":  0.,
        "pa_aper":   0.}

    for key in keylist.keys():
        info[key] = hdr.get(key, default=keylist[key])

    # For FUV, the keyword for exposure time depends on segment.
    exptime_key = cosutil.segmentSpecificKeyword("exptime", info["segment"])
    exptime_default = hdr.get("exptime", default=0.)
    info["exptime"] = hdr.get(exptime_key, default=exptime_default)

    # FUV detector high voltage level (commanded, raw)
    hvlevel_key = cosutil.segmentSpecificKeyword("hvlevel", info["segment"])
    info["hvlevel"] = hdr.get(hvlevel_key, default=NOT_APPLICABLE)

    # Copy exptime to orig_exptime, so we can modify exptime but save the
    # original value.  NOTE:  for TIME-TAG data this value will be replaced
    # with the difference between the first and last values in the TIME column.
    info["orig_exptime"] = info["exptime"]

    if info["tagflash"] and info["numflash"] < 1:
        info["tagflash"] = False

    # Reset the subarray flag if the "subarray" is the entire detector.
    if info["subarray"]:
        if info["detector"] == "FUV":
            # Indices 0, 1, 2, 3 are for FUVA, while 4, 5, 6, 7 are for FUVB.
            if info["segment"] == "FUVA":
                sub_number = "0"
            else:
                sub_number = "4"
        else:
            sub_number = "0"
        xsize = hdr.get("size"+sub_number+"x", default=0)
        ysize = hdr.get("size"+sub_number+"y", default=0)
        if info["detector"] == "FUV" and xsize == FUV_X and ysize == FUV_Y:
            info["subarray"] = False
        elif xsize == NUV_X and ysize == NUV_Y:
            info["subarray"] = False

    return info

def getSwitchValues(phdr):
    """Get calibration switch values from the primary header.

    The input argument phdr is the primary header, as provided by the fits
    module.  The function value is a dictionary of keyword = value pairs.
    Note that the keyword values will be converted to upper case.  If a
    keyword is missing from the header, it will still be included in the
    dictionary, but its value will be set to the NOT_APPLICABLE string.

    Parameters
    ----------
    phdr: ``astropy.io.fits.header.Header`` object
        Primary header.

    Returns
    -------
    switches: dictionary
        Calibration switch keywords and their values from the primary
        header of the input file.
    """

    switches = {}

    for key in ["dqicorr", "randcorr", "tempcorr", "geocorr", "igeocorr",
                "dgeocorr", "xwlkcorr", "ywlkcorr", "trcecorr", "algncorr",
                "deadcorr", "flatcorr", "doppcorr", "helcorr", "phacorr",
                "brstcorr", "badtcorr", "x1dcorr", "wavecorr", "backcorr",
                "fluxcorr", "photcorr", "tdscorr", "hvdscorr", "statflag"]:
        switches[key]  = cosutil.getSwitch(phdr, key)

    return switches

def getRefFileNames(phdr):
    """Get reference file names from the primary header.

    The input argument phdr is the primary header, as provided by the astropy fits
    module.  The function value is a dictionary of keyword = value pairs.
    If a keyword is missing from the header, it will still be included in
    the dictionary, but its value will be set to the NOT_APPLICABLE string.
    If the name includes an environment variable (Unix-style or IRAF-style),
    the name will be expanded to a complete pathname.  Keys of the form
    "bpixtab_hdr" (for example) are the values read directly from the
    header, while keys of the form "bpixtab" have been translated to full
    path names (operating system dependent).

    Parameters
    ----------
    phdr: ``astropy.io.fits.header.Header`` object
        Primary header.

    Returns
    -------
    reffiles: dictionary
        Reference file keywords and their values from the primary
        header of the input file.
    """

    reffiles = {}

    for key in ["flatfile", "hvtab", "xwlkfile", "ywlkfile",
                "bpixtab", "gsagtab", "spottab", "brftab",
                "geofile", "dgeofile", "twozxtab", "deadtab", "phafile",
                "phatab", "brsttab", "badttab", "tracetab",
                "xtractab", "lamptab", "disptab", "fluxtab",
                "imphttab", "phottab", "spwcstab", "wcptab",
                "tdstab", "proftab", "hvdstab"]:
        reffiles[key+"_hdr"] = phdr.get(key, default=NOT_APPLICABLE)
        reffiles[key] = cosutil.expandFileName(reffiles[key+"_hdr"])

    if phdr["obstype"] == "SPECTROSCOPIC":
        if phdr.get("fluxtab", "missing") == "missing":
            reffiles["fluxtab"] = reffiles["phottab"]
            reffiles["fluxtab_hdr"] = reffiles["phottab_hdr"]
    else:
        if phdr.get("imphttab", "missing") == "missing":
            reffiles["imphttab"] = reffiles["phottab"]
            reffiles["imphttab_hdr"] = reffiles["phottab_hdr"]

    return reffiles

def resetSwitches(switches, reffiles):
    """Reset calibration switches if required reference file is "N/A".

    If a calibration step needs one or more reference files, and if the
    name of any such file is given in the header as "N/A", the calibration
    step cannot be done.  This function checks some steps and resets the
    switch from PERFORM to SKIPPED if a required reference file is "N/A".

    Parameters
    ----------
    switches: dictionary
        Keyword and value for calibration switches.

    reffiles: dictionary
        Keyword and value for reference file names.
    """

    check_these = {"badtcorr": ["badttab"],
                   "xwlkcorr": ["xwlkfile"],
                   "ywlkcorr": ["ywlkfile"],
                   "tdscorr": ["tdstab"]}
    #check_these = {"badtcorr": ["badttab"],
    #               "brstcorr": ["brsttab"],
    #               "dqicorr": ["bpixtab"],
    #               "flatcorr": ["flatfile"],
    #               "deadcorr": ["deadtab"],
    #               "geocorr": ["geofile"],
    #               "dgeocorr": ["dgeofile"],
    #               "x1dcorr": ["xtractab", "disptab"],
    #               "fluxcorr": ["fluxtab"],
    #               "photcorr": ["imphttab"]}

    for switch_key in check_these.keys():
        not_specified = []
        if switches[switch_key] == "PERFORM":
            for reffile_key in check_these[switch_key]:
                if reffiles[reffile_key] == NOT_APPLICABLE:
                    not_specified.append(reffile_key)
        if not_specified:
            switches[switch_key] = "SKIPPED"
            cosutil.printWarning("%s will be set to SKIPPED because" %
                                 switch_key.upper())
            for (i, reffile_key) in enumerate(not_specified):
                keyword = reffile_key.upper()
                if i == 0:
                    message = "%s = %s" % (keyword, NOT_APPLICABLE)
                else:
                    message += ", %s = %s" % (keyword, NOT_APPLICABLE)
            cosutil.printContinuation(message)



import math
import os
import numpy as np
from scipy.signal import windows as scipysignal
from scipy import ndimage
from astropy.io import fits
from .calcosparam import *
from . import ccos
from . import cosutil
from . import findshift1
from . import shiftfile

# Each element of wavecal_info is a dictionary with the keys:
# "time", "cenwave", "fpoffset", "shift_dict", "rootname", "filename".
# The variable name wc_dict is used for one element of wavecal_info.

def findWavecalShift(input, shift_file, info, wcp_info):
    """Find the shift from a wavecal image.

    Note that the input file will be opened read-write, and several
    keywords (e.g. SHIFT1A, SHIFT1B, and SHIFT1C if NUV) will be updated
    in the header.

    Parameters
    ----------
    input: str
        Name of an x1d FITS file containing a wavecal observation.

    shift_file: str
        Name of a user-supplied file to override shifts, or None.

    info: dictionary
        Keywords and values from the headers of the input file.

    wcp_info: array_like
        Data (one row) from the wavecal parameters table.

    Returns
    -------
    (shift_dict, fp_dict)
        shift_dict is a dictionary or None.  Keys are header keywords (but
        lower case) for wavecal shift information, values are the values
        for updating the header.  shift_dict will be None if there is no
        data in the first extension of input.
        fp_dict is a dictionary.  Keys are tuples of segment or stripe
        name (upper case) and fpoffset.  Values are fp_pixel_shift.  This
        is only needed for the case that there is a science exposure with
        no wavecal taken at the same fpoffset, but there is a wavecal for
        at least one fpoffset.
    """

    # Note that we open the x1d table read-write, so we can update keywords.
    fd = fits.open(input, mode="update")
    phdr = fd[0].header
    sci_extn = fd["SCI"]
    if sci_extn.data is None or len(sci_extn.data) == 0:
        fd.close()
        return None

    reffiles = {}
    lamptab = cosutil.expandFileName(phdr["lamptab"])
    reffiles["lamptab"] = lamptab

    wcptab = cosutil.expandFileName(phdr["wcptab"])
    reffiles["wcptab"] = wcptab

    x_offset = sci_extn.header.get("x_offset", default=0)
    info["x_offset"] = x_offset

    nrows = sci_extn.data.shape[0]
    gross = sci_extn.data.field("gross")
    nelem = len(gross[0])
    exptime = sci_extn.data.field("exptime")
    segment = sci_extn.data.field("segment")

    detector = info["detector"]
    xc_range = wcp_info.field("xc_range")
    stepsize = wcp_info.field("stepsize")
    try:
        search_offset = wcp_info.field("search_offset")
    except KeyError:
        search_offset = 0.

    # Replace shift values in segment B with values from segment A?
    if "FUVB" in segment:
        override_segment_B = cosutil.checkForNoWavecalData(
                info["opt_elem"], info["cenwave"], "FUVB", lamptab)
    else:
        override_segment_B = False
    segment_A_present = ("FUVA" in segment)

    # segment will be added to the filter in the loop below.
    filter = {"opt_elem": info["opt_elem"],
              "cenwave": info["cenwave"]}
    # If the FPOFFSET column is present in the lamptab, include fpoffset
    # in the filter.  If not, use an offset when doing the cross correlation.
    if cosutil.findColumn(lamptab, "fpoffset"):
        filter["fpoffset"] = info["fpoffset"]
    # initial_offset is used as the center of the search range, when
    # matching the spectrum to the template.
    got_pixel_shift = cosutil.findColumn(lamptab, "fp_pixel_shift")
    if got_pixel_shift:
        initial_offset = search_offset
    else:
        initial_offset = info["fpoffset"] * stepsize + search_offset

    shift_dict = {}

    index = segment.argsort()
    save_spectra = {}
    save_templates = {}
    fp_pixel_shift = {}
    fp_dict = {}

    for row in index:
        filter["segment"] = segment[row]
        lamp_info = cosutil.getTable(lamptab, filter)
        if lamp_info is None:
            raise MissingRowError("Missing row in LAMPTAB; filter = %s" %
                                  str(filter))
        # Save gross, but convert from count rate back to counts.
        save_spectra[segment[row]] = gross[row] * exptime[row]
        raw_template = lamp_info.field("intensity")[0]
        save_templates[segment[row]] = \
                cosutil.getTemplate(raw_template, x_offset, nelem)
        if got_pixel_shift:
            fp_pixel_shift[segment[row]] = \
                    lamp_info.field("fp_pixel_shift")[0]
        else:
            fp_pixel_shift[segment[row]] = 0.
        # Add elements to fp_dict, one for each fpoffset.
        readFpPixelShift(info, lamptab, segment[row], stepsize, fp_dict)

    # find offset in dispersion direction
    fs1 = findshift1.Shift1(save_spectra, save_templates, info, reffiles,
                            xc_range, fp_pixel_shift, initial_offset)
    fs1.findShifts()

    if override_segment_B:
        if segment_A_present:           # copy shift1 from A to B
            fs1.setShift1("FUVB", fs1.getShift1("FUVA"))
        else:
            fs1.setShift1("FUVB", 0., fp=False)

    # Did the user supply a file with overrides for the shifts?
    if shift_file is None:
        user_shifts = None
    else:
        user_shifts = shiftfile.ShiftFile(shift_file,
                                          info["root"], info["fpoffset"])

    if detector == "FUV":
        cosutil.printMsg("segment   shift    err  [orig.]    FP   chi sq (n)",
                         VERBOSE)
        cosutil.printMsg("-------  --------- ---- ------- ------  ----------",
                         VERBOSE)
    else:
        cosutil.printMsg("stripe    shift    err  [orig.]    FP   chi sq (n)",
                         VERBOSE)
        cosutil.printMsg("------   --------- ---- ------- ------  ----------",
                         VERBOSE)

    # Print and save results.
    for row in index:

        # This is the offset in the cross-dispersion direction.
        key = "shift2" + segment[row][-1].lower()
        if override_segment_B and segment[row] == "FUVB":
            if segment_A_present:       # copy shift2 from A to B
                shift_dict[key] = sci_extn.header.get("shift2a", default=0.)
            else:
                shift_dict[key] = 0.
        else:
            shift_dict[key] = sci_extn.header.get(key, default=0.)
        sci_extn.header[key] = shift_dict[key]

        # Zero for dpixel1[a-c] is appropriate for auto/GO wavecals.
        key = "dpixel1" + segment[row][-1].lower()
        sci_extn.header[key] = 0.
        shift_dict[key] = 0.

        key = "shift1" + segment[row][-1].lower()
        user_specified = False
        if user_shifts is not None:
            ((user_shift1, user_shift2), nfound) = \
                        user_shifts.getShifts(("any", segment[row]))
            if user_shift1 is not None:
                fs1.setShift1(segment[row], user_shift1)
                user_specified = True
        shift_segment = fs1.getShift1(segment[row])
        scatter = fs1.getScatter(segment[row])
        measured_shift1 = fs1.getMeasuredShift1(segment[row])
        fp_pixel_shift_seg = fs1.getFpPixelShift(segment[row])
        sci_extn.header[key] = shift_segment
        shift_dict[key] = shift_segment

        message = " %4s    %9.4f %4.2f [%5.1f] %6.1f  %6.1f (%d)" % \
                (segment[row], shift_segment, scatter, measured_shift1,
                 fp_pixel_shift_seg,
                 fs1.getChiSq(segment[row]), fs1.getNdf(segment[row]))
        if user_specified:
            message = message + "  # user-specified"
        elif override_segment_B and segment[row] == "FUVB":
            if segment_A_present:
                message = message + "  # based on FUVA value"
            else:
                message = message + "  # set to default (no FUVA)"
        elif not fs1.getSpecFound(segment[row]):
            message = message + "  # not found"
        cosutil.printMsg(message, VERBOSE)

        # Save Chi square and the number of degrees of freedom in the
        # shift dictionary, and update the keywords in the x1d header.
        key = "chi_sq_" + segment[row][-1].lower()
        shift_dict[key] = round(fs1.getChiSq(segment[row]), 1)
        sci_extn.header[key] = shift_dict[key]
        key = "ndf_" + segment[row][-1].lower()
        shift_dict[key] = fs1.getNdf(segment[row])
        sci_extn.header[key] = shift_dict[key]

    fd.close()

    return (shift_dict, fp_dict)

def readFpPixelShift(info, lamptab, segment, stepsize, fp_dict):
    """Read all four fp_pixel_shift values from the lamptab.

    Parameters
    ----------
    info: dictionary
        Keywords and values from the headers of the input file.

    lamptab: str
        Name of the template lamp table.

    segment: str
        Current segment or stripe name.

    stepsize: int
        Offset in pixels corresponding to one step of the OSM.  This is
        only used if the lamptab does not have FPOFFSET and FP_PIXEL_SHIFT
        columns.

    fp_dict: dictionary
        Updated in-place.  The keys are tuples of (segment, fpoffset),
        the segment or stripe name (upper case) and the integer offset
        (-2, -1, 0, 1) of the OSM.  Entries for the current segment and
        all fpoffset in the lamptab will be added to fp_dict.
    """

    # We expect to get four rows matching this filter, one row for each
    # fpoffset.
    filter = {"opt_elem": info["opt_elem"],
              "cenwave": info["cenwave"],
              "segment": segment}
    lamp_info = cosutil.getTable(lamptab, filter)
    if lamp_info is None:
        raise MissingRowError("Missing row in LAMPTAB; filter = %s" %
                              str(filter))

    names = []
    for name in lamp_info.names:
        names.append(name.lower())
    if "fpoffset" in names and "fp_pixel_shift" in names:
        fpoffset = lamp_info.field("fpoffset")
        fp_pixel_shift = lamp_info.field("fp_pixel_shift")
        for i in range(len(fpoffset)):
            fp_dict[(segment, fpoffset[i])] = fp_pixel_shift[i]
    else:
        for i in [-2, -1, 0, 1]:
            fp_dict[(segment, i)] = float(i * stepsize)

def storeWavecalInfo(wavecal_info, time, cenwave, fpoffset,
                     shift_dict, fp_dict,
                     rootname, filename):
    """Append the current info to the wavecal_info list.

    shift_dict can have any of the following keys:
        "shift1a" for FUV segment A or NUV stripe A,
        "shift1b" for FUV segment B or NUV stripe B,
        "shift1c" for NUV stripe C,
        "shift2a", "shift2b", "shift2c",
        "dpixel1a", "dpixel1b", "dpixel1c",
        "chi_sq_a", "chi_sq_b", "chi_sq_c",
        "ndf_a", "ndf_b", "ndf_c".
    shift1[abc] is the measured shift in the dispersion direction, and
    shift2[abc] is the measured shift in the cross-dispersion direction.

    fp_dict has (segment, fpoffset) as key.  segment is either a segment
    or stripe name (upper case), and fpoffset is an int, -2, -1, 0 or 1.
    The values are fp_pixel_shift (from that column in the lamptab).

    The input information will be combined into a dictionary, which will
    then be appended to wavecal_info.  wavecal_info will be sorted in
    increasing order of time.

    Parameters
    ----------
    wavecal_info: list of wavecal information dictionaries
        The other arguments will be used to create a wavecal information
        dictionary, and that will be appended to wavecal_info.
        A wavecal information dictionary has keys "time", "fpoffset",
        "shift_dict", "rootname", and "filename".  The values are the
        arguments with those names, except that if filename includes a
        directory, that will be removed before saving in the dictionary.

    time: float
        Time of observation, MJD at middle of exposure

    cenwave: int
        Central wavelength, used to select entries from wavecal_info

    fpoffset: int
        OSM position, used to select entries from wavecal_info

    shift_dict: dictionary
        A dictionary of keyword names and shifts

    fp_dict: dictionary
        Dictionary of fp_pixel_shift for each key (segment, fpoffset).

    rootname: str
        The rootname (typically lower case) of the observation

    filename: str
        The name of the raw file, including directory; for FUV this will
        typically be the _a.fits name
    """

    wc_dict = {}
    wc_dict["time"]       = time
    wc_dict["cenwave"]    = cenwave
    wc_dict["fpoffset"]   = fpoffset
    wc_dict["shift_dict"] = shift_dict
    wc_dict["fp_dict"]    = fp_dict
    wc_dict["rootname"]   = rootname
    wc_dict["filename"]   = os.path.basename(filename)

    wavecal_info.append(wc_dict)
    if len(wavecal_info) > 1:
        wavecal_info.sort(key=keyTime)

def keyTime(wc_dict):
    """Return the time in a wavecal_info entry.

    Parameters
    ----------
    wc_dict: dictionary
        A wavecal information dictionary (one element of wavecal_info)

    Returns
    -------
    float
        The element of wc_dict with key "time"
    """

    return wc_dict["time"]

def returnWavecalShift(wavecal_info, wcp_info, cenwave, fpoffset, time):
    """Return the matching shift dictionary from wavecal_info.

    The element of wavecal_info that matches cenwave and fpoffset will be
    extracted.  If there are multiple entries that match, we'll find the
    two that are closest to the time of the observation and linearly
    interpolate the shifts at that time.

    If there is no entry in wavecal_info for the current cenwave and
    fpoffset, we'll find the entry that is closest in time to the time of
    the observation.  If the difference in time for that entry is not too
    large (based on info from the wavecal parameters table), we'll use that
    entry and correct the shift in the dispersion direction by adding the
    difference in FP_PIXEL_SHIFT values (from the lamptab).

    None will be returned if wavecal_info is empty.

    Parameters
    ----------
    wavecal_info: list of dictionaries
        List of wavecal information dictionaries.

    wcp_info: astropy.io.fits record object
        Data (one row) from the wavecal parameters table.

    cenwave: int
        Central wavelength, used to select entries from wavecal_info.

    fpoffset: int
        OSM position for the current science exposure, used to select
        entries from wavecal_info.

    time: float
        Time of observation, MJD at middle of exposure.

    Returns
    -------
    tuple, or None
        A pair of dictionaries and a string.  For the dictionaries, the
        key is the keyword name for the shift (shift1a, shift1b, shift1c,
        shift2a, shift2b, shift2c).  For the first dictionary, the value
        is the shift at the specified time.  For the second dictionary,
        the value is the slope (pixels/s).  The string is the name or
        names of the wavecal files (separated by a blank, if there are
        two) that were used to find the shifts.
    """

    if len(wavecal_info) < 1:
        return None

    slope_dict = {}             # initial value

    # Extract those elements of wavecal_info that match cenwave and fpoffset.
    subset_wavecal_info = selectWavecalInfo(wavecal_info, cenwave, fpoffset)

    if len(subset_wavecal_info) == 1:
        shift_dict = subset_wavecal_info[0]["shift_dict"]
        filename = subset_wavecal_info[0]["filename"]
    elif len(subset_wavecal_info) > 1:
        (shift_dict, slope_dict, filename) = \
        interpolateWavecal(subset_wavecal_info, time)
    else:
        # No matching row; find nearest in time.
        wc_dict = minTimeWavecalInfo(wavecal_info, time, cenwave,
                                     wcp_info.field("max_time_diff"))
        if wc_dict is None or cenwave != wc_dict["cenwave"]:
            cosutil.printWarning(
                    "No matching wavecal info; zero shift assumed.")
            shift_dict = None
        else:
            # No matching element in wavecal_info, but wc_dict should
            # match everything except fpoffset.  Apply a correction from
            # the wavecal fpoffset to the science exposure fpoffset.
            # fpoffset is for the current science exposure.
            # wc_dict["fpoffset"] is for the closest wavecal exposure.
            filename = wc_dict["filename"]
            shift_dict = wc_dict["shift_dict"].copy()
            fp_dict = wc_dict["fp_dict"]
            # Get any segment in fp_dict, use to construct segment in loop.
            fp_dict_keys = list(fp_dict.keys())
            (segment, fpoff) = fp_dict_keys.pop()       # ignore fpoff
            # Apply correction to all shift1[abc] values.
            for shift_dict_key in shift_dict:
                if not shift_dict_key.startswith("shift1"):
                    continue
                letter = shift_dict_key[-1].upper()
                key1 = (segment[:-1] + letter, fpoffset)
                key2 = (segment[:-1] + letter, wc_dict["fpoffset"])
                if key1 not in fp_dict or key2 not in fp_dict:
                    raise RuntimeError("fpoffset = %d or %d not found"
                                       " in the lamptab." %
                                       (fpoffset, wc_dict["fpoffset"]))
                # This is a difference in fp_pixel_shift values.
                correction = fp_dict[key1] - fp_dict[key2]
                shift_dict[shift_dict_key] += correction
                cosutil.printMsg("Info:  Keyword %s adjusted by adding %.6g" %
                                 (shift_dict_key, correction), VERBOSE)

    if shift_dict is None:
        return None

    # The default is zero slope.
    if not slope_dict:
        for key in shift_dict.keys():
            slope_dict[key] = 0.

    return (shift_dict, slope_dict, filename)

def returnExactMatch(wavecal_info, rootname):
    """Return the shift dictionary for the specified wavecal rootname.

    The rootname of a wavecal exposure is used to extract one element of
    wavecal_info (there should always be exactly one matching element).
    The shift dictionary from that matchine element is then returned.

    None will be returned if wavecal_info is empty.

    Parameters
    ----------
    wavecal_info: list of dictionaries
        List of wavecal information dictionaries

    rootname: str
        Used to find the appropriate element of wavecal_info

    Returns
    -------
    dictionary or None
        The shift dictionary corresponding to rootname, or None
    """

    if len(wavecal_info) < 1:
        return None

    for wc_dict in wavecal_info:
        if wc_dict["rootname"] == rootname:
            return wc_dict["shift_dict"]

    # We shouldn't get here.
    raise RuntimeError("There should have been a matching element.")

def selectWavecalInfo(wavecal_info, cenwave, fpoffset):
    """Return a list of all matching elements of wavecal_info.

    Parameters
    ----------
    wavecal_info: list of dictionaries
        List of wavecal information dictionaries

    cenwave: int
        Central wavelength, used to select entries from wavecal_info

    fpoffset: int
        Used to find one or more elements of wavecal_info

    Returns
    -------
    list
        List of dictionaries in wavecal_info that match cenwave and
        fpoffset
    """

    subset_wavecal_info = []

    for wc_dict in wavecal_info:
        if wc_dict["cenwave"] == cenwave and wc_dict["fpoffset"] == fpoffset:
            subset_wavecal_info.append(wc_dict)

    return subset_wavecal_info

def minTimeWavecalInfo(wavecal_info, time, cenwave, max_time_diff):
    """Return the element of wavecal_info that is closest to time.

    The element of wavecal_info that is closest in time to the specified
    time will be selected.  If the difference between the time for that
    element and the specified time is less than max_time_diff, that element
    of wavecal_info will be returned; otherwise, None will be returned.

    Associations can include exposures with different values of central
    wavelength, but it would not make sense to use a wavecal taken with
    one value of cenwave for a science exposure taken with a different
    value of cenwave, so elements of wavecal_info that do not match
    cenwave are ignored.

    Parameters
    ----------
    wavecal_info: list of dictionaries
        List of wavecal information dictionaries

    time: float
        Time of a science observation (MJD at middle of exposure)

    cenwave: int
        Central wavelength of the current science exposure, used to reject
        non-matching entries from wavecal_info

    max_time_diff: float
        Cutoff for time difference between time and a wavecal observation

    Returns
    -------
    dictionary
        The shift dictionary closest in time to time, or None if there
        is none that is within max_time_diff
    """

    index = -1
    for i in range(len(wavecal_info)):
        wc_dict = wavecal_info[i]
        if wc_dict["cenwave"] != cenwave:
            continue
        delta_t = abs(time - wc_dict["time"])
        if index < 0 or delta_t < min_time:
            index = i
            min_time = delta_t

    if index < 0:
        wc_dict = None
    else:
        wc_dict = wavecal_info[index]
        if min_time > max_time_diff:
            wc_dict = None

    return wc_dict

def interpolateWavecal(wavecal_info, time):
    """Interpolate to get a shift dictionary at the specified time.

    wavecal_info is assumed to be sorted in increasing order of time.
    If the time of observation is earlier than the first entry or later
    than the last entry in wavecal_info, then the shift dictionary for the
    first or last element respectively of wavecal_info will be returned.
    Otherwise, the pair of entries that bracket the time of observation
    will be selected, and the shifts in the shift dictionaries will be
    linearly interpolated at the time of observation.

    None will be returned if wavecal_info is empty.

    Parameters
    ----------
    wavecal_info: list of dictionaries
        list of wavecal information dictionaries

    time: float
        time of observation, MJD at middle of exposure

    Returns
    -------
    tuple, or None
        A pair of dictionaries and a string.  For the dictionaries, the
        key is the keyword name for the shift (shift1a, shift1b, shift1c,
        shift2a, shift2b, shift2c).  For the first dictionary, the value is
        the shift at the specified time.  For the second dictionary, the
        value is the slope in pixels per second.  The string is the name or
        names of the wavecal files (separated by a blank, if there are two)
        that were used to find the shifts.
    """

    wlen = len(wavecal_info)
    if wlen < 1:
        return None

    slope_dict = {}
    for key in wavecal_info[0]["shift_dict"]:
        slope_dict[key] = 0.

    if time <= wavecal_info[0]["time"]:
        return (wavecal_info[0]["shift_dict"], slope_dict,
                wavecal_info[0]["filename"])

    if time >= wavecal_info[wlen-1]["time"]:
        return (wavecal_info[wlen-1]["shift_dict"], slope_dict,
                wavecal_info[wlen-1]["filename"])

    for i in range(1, wlen):
        wc_i = wavecal_info[i]
        if time <= wavecal_info[i]["time"]:
            t0 = wavecal_info[i-1]["time"]
            t1 = wavecal_info[i]["time"]
            shift_dict0 = wavecal_info[i-1]["shift_dict"].copy()
            filename = wavecal_info[i-1]["filename"]
            if t0 != t1:
                filename = filename + " " + wavecal_info[i]["filename"]
                shift_dict1 = wavecal_info[i]["shift_dict"]
                # for each segment or stripe ...
                for key in shift_dict0.keys():
                    shift0 = shift_dict0[key]
                    if key in shift_dict1:
                        shift1 = shift_dict1[key]
                        p = (time - t0) / (t1 - t0)
                        q = 1. - p
                        shift = q * shift0 + p * shift1
                        shift_dict0[key] = shift
                        slope_dict[key] = (shift1 - shift0) \
                                          / ((t1 - t0) * SEC_PER_DAY)

            return (shift_dict0, slope_dict, filename)

def findWavecalSpectrum(corrtag, info, reffiles):
    """Find the offset of a wavecal spectrum in the cross-dispersion direction.

    Note that it is assumed that all wavecals are taken in time-tag mode.

    Parameters
    ----------
    corrtag: str
        Name of the corrtag FITS file containing a wavecal.

    info: dictionary
        Keywords and values from the headers of the input file.

    reffiles: dictionary
        Reference file names.

    Returns
    -------
    (shift2, xd_shifts, xd_locns, lamp_is_on): tuple of a float, two
    dictionaries, and a boolean flag
        shift2 is the offset (average of those found, if NUV) from
        nominal in the cross-dispersion direction, in pixels; this value
        will be zero if the offset could not be determined.  Dictionaries
        xd_shifts and xd_locns use the segment or stripe name as the
        key; the value for xd_shifts is the shift from nominal, and the
        value for xd_locns is the location where the spectrum was found
        (projected onto the left edge).  lamp_is_on is a flag that
        indicates whether the lamp was actually on.
    """

    fd = fits.open(corrtag, mode="copyonwrite")
    phdr = fd[0].header
    sci_extn = fd["EVENTS"]
    if sci_extn.data is None or len(sci_extn.data) == 0:
        fd.close()
        return (0., {}, {}, False)

    xtractab = reffiles["xtractab"]

    wcp_info = cosutil.getTable(reffiles["wcptab"],
                                filter={"opt_elem": info["opt_elem"]},
                                exactly_one=True)
    xd_range = wcp_info.field("xd_range")[0]
    box = wcp_info.field("box")[0]

    if info["detector"] == "FUV":
        xi = sci_extn.data.field("XCORR")
        eta = sci_extn.data.field("YCORR")
    else:
        xi = sci_extn.data.field("RAWX")
        eta = sci_extn.data.field("RAWY")

    dq = sci_extn.data.field("DQ")

    (shift2, xd_shifts, xd_locns, lamp_is_on) = \
    ttFindWavecalSpectrum(xi, eta, dq, info, xd_range, box, xtractab)

    fd.close()

    cosutil.printMsg("Shift (location) in cross-dispersion direction:")
    # keys = xd_shifts.keys()
    # keys.sort()
    keys = sorted(xd_shifts)
    for key in keys:
        if xd_shifts[key] is None:
            cosutil.printMsg("%4s    ---- (%5.1f)  # not found in XD" \
                             % (key, xd_locns[key]))
        else:
            cosutil.printMsg("%4s  %6.1f (%5.1f)" % \
                             (key, xd_shifts[key], xd_locns[key]))
    cosutil.printMsg("  avg %6.1f" % shift2)

    return (shift2, xd_shifts, xd_locns, lamp_is_on)

def ttFindWavecalSpectrum(xi, eta, dq, info, xd_range, box, xtractab):
    """Find the offset of a wavecal spectrum in cross-dispersion direction.

    Parameters
    ----------
    xi: array_like
        Corrected pixel coordinates in the dispersion direction.

    eta: array_like
        Corrected pixel coordinates in the cross-dispersion direction.

    dq: array_like
        Data quality flags.

    info: dictionary
        Header keywords and values.

    xd_range: int
        Search within + or - xd_range from the nominal location for the
        peak in the cross-dispersion direction.

    box: int
        Smooth the cross-dispersion profile with a box of this width before
        looking for the maximum.

    xtractab: str
        Name of the 1-D extraction parameters table.

    Returns
    -------
    (shift2, xd_shifts, xd_locns, lamp_is_on):  tuple of a float, two
    dictionaries, and a boolean flag
        shift2 is the offset (average of those found, if NUV) from
        nominal in the cross-dispersion direction, in pixels; this value
        will be zero if the offset could not be determined.  Dictionaries
        xd_shifts and xd_locns use the segment or stripe name as the
        key; the value for xd_shifts is the shift from nominal, and the
        value for xd_locns is the location where the spectrum was found
        (projected onto the left edge).  lamp_is_on is a flag that
        indicates whether the lamp was actually on.
    """

    if len(xi) < 1:
        shift2 = 0.
        xd_shifts = {}
        xd_locns = {}
        lamp_is_on = False
        if info["detector"] == "FUV":
            segment_list = [info["segment"]]
        elif info["obstype"] == "IMAGING":
            segment_list = ["NUVA"]
        else:
            segment_list = ["NUVA", "NUVB", "NUVC"]
        for segment in segment_list:
            xd_shifts[segment] = None
            xd_locns[segment] = 0.
        return (shift2, xd_shifts, xd_locns, lamp_is_on)

    filter = {"segment": info["segment"],
              "opt_elem": info["opt_elem"],
              "cenwave": info["cenwave"],
              "aperture": "WCA"}

    if info["detector"] == "FUV":
        (shift2, xd_shifts, xd_locns) = ttFindFUV(xi, eta, dq,
                info["life_adj_offset"], xd_range, box, filter, xtractab)
    elif info["obstype"] == "IMAGING":
        (shift2, xd_shifts, xd_locns) = ttFindImagingWavecal(xi, eta, dq,
                info["life_adj_offset"], xd_range, box, filter, xtractab)
    else:
        (shift2, xd_shifts, xd_locns) = ttFindNUV(xi, eta, dq,
                info["life_adj_offset"], xd_range, box, filter, xtractab)

    if shift2 is None:
        shift2 = 0.

    # Was the lamp was actually on?  shift2 would have been None if we
    # didn't find the shift, but this test is more quantitative and prints
    # some statistics.
    lamp_is_on = cosutil.isLampOn(xi, eta, dq, info, xtractab, shift2)

    return (shift2, xd_shifts, xd_locns, lamp_is_on)

def ttFindFUV(xi, eta, dq, life_adj_offset, xd_range, box, filter, xtractab):

    xdisp = np.zeros(FUV_Y, dtype=np.float32)

    shift2 = None
    xd_shifts = {}
    xd_locns = {}
    xtract_info = cosutil.getTable(xtractab, filter, exactly_one=True)
    if xtract_info is not None:
        slope = xtract_info.field("slope")[0]
        # Collapse the data along the dispersion direction, putting the
        # result in xdisp.
        ccos.xy_collapse(xi, eta, dq, slope, xdisp)
        (shift2, y) = ttFindSpec(xdisp, xtract_info,
                                 life_adj_offset, xd_range, box)
        segment = filter["segment"]
        xd_shifts = {segment: shift2}
        xd_locns = {segment: y}

    return (shift2, xd_shifts, xd_locns)

def ttFindImagingWavecal(xi, eta, dq, life_adj_offset, xd_range,
                         box, filter, xtractab):

    xdisp = np.zeros(NUV_Y, dtype=np.float32)

    xd_shifts = {}
    xd_locns = {}

    filter["segment"] = "NUVA"
    xtract_info = cosutil.getTable(xtractab, filter)
    if xtract_info is not None:
        slope = xtract_info.field("slope")[0]
        # Collapse the data along the dispersion direction.
        ccos.xy_collapse(xi, eta, dq, slope, xdisp)
        (shift2, y) = ttFindSpec(xdisp, xtract_info,
                                 life_adj_offset, xd_range, box)
        xd_shifts["NUVA"] = shift2
        xd_locns["NUVA"] = y

    return (shift2, xd_shifts, xd_locns)

def ttFindNUV(xi, eta, dq, life_adj_offset, xd_range, box, filter, xtractab):

    xdisp = np.zeros(NUV_Y, dtype=np.float32)

    xd_shifts = {}
    xd_locns = {}
    got_shift = False

    filter["segment"] = "NUVA"
    xtract_info = cosutil.getTable(xtractab, filter, exactly_one=True)
    if xtract_info is not None:
        slope = xtract_info.field("slope")[0]
        outlier_limit = xtract_info.field("height")[0] / 4.
        # Collapse the data along the dispersion direction.
        ccos.xy_collapse(xi, eta, dq, slope, xdisp)
        (shift2, y) = ttFindSpec(xdisp, xtract_info,
                                 life_adj_offset, xd_range, box)
        xd_shifts["NUVA"] = shift2
        xd_locns["NUVA"] = y
        if shift2 is not None:
            got_shift = True

    filter["segment"] = "NUVB"
    xtract_info = cosutil.getTable(xtractab, filter, exactly_one=True)
    if xtract_info is not None:
        slope = xtract_info.field("slope")[0]
        outlier_limit = xtract_info.field("height")[0] / 4.
        ccos.xy_collapse(xi, eta, dq, slope, xdisp)
        (shift2, y) = ttFindSpec(xdisp, xtract_info,
                                 life_adj_offset, xd_range, box)
        xd_shifts["NUVB"] = shift2
        xd_locns["NUVB"] = y
        if shift2 is not None:
            got_shift = True

    filter["segment"] = "NUVC"
    xtract_info = cosutil.getTable(xtractab, filter, exactly_one=True)
    if xtract_info is not None:
        slope = xtract_info.field("slope")[0]
        outlier_limit = xtract_info.field("height")[0] / 4.
        ccos.xy_collapse(xi, eta, dq, slope, xdisp)
        (shift2, y) = ttFindSpec(xdisp, xtract_info,
                                 life_adj_offset, xd_range, box)
        xd_shifts["NUVC"] = shift2
        xd_locns["NUVC"] = y
        if shift2 is not None:
            got_shift = True

    if got_shift:
        # Find the average shift, ignoring outliers (beyond 1/4 the
        # extraction height).
        shifts = []
        for shift in xd_shifts.values():
            if shift is not None:
                shifts.append(shift)
        shifts.sort()
        median_shift = shifts[len(shifts)//2]
        sum_s = 0.
        nsum = 0.
        for shift in shifts:
            if abs(shift - median_shift) < outlier_limit:
                sum_s += shift
                nsum += 1.
        shift2 = sum_s / nsum
    else:
        shift2 = None

    return (shift2, xd_shifts, xd_locns)

def ttFindSpec(xdisp, xtract_info, life_adj_offset, xd_range, box):
    """Find the location in the cross-dispersion direction.

    Parameters
    ----------
    xdisp: array_like
        The cross-dispersion profile, 1-D array of time-tag data collapsed
        along the dispersion axis, but taking into account the tilt of the
        spectrum.

    xtract_info: array_like
        Data block (but just one row) from the xtractab.

    life_adj_offset: float
        Normally this will be 0.  If the LIFE_ADJ keyword is -1, however,
        indicating that the aperture block is not at one of the recognized
        "lifetime positions," life_adj_offset will be the expected offset
        (in pixels) of the wavecal spectrum from lifetime position 1.

    xd_range: int
        Search within + or - xd_range from the nominal location for the
        peak in xdisp.

    box: int
        Smooth xdisp with a box of this width before looking for the
        maximum.

    Returns
    -------
    (shift2, y): tuple of two floats
        shift2 is the shift from nominal in the cross-dispersion
        direction (or None), and y is the location of the spectrum.
        The location is based on fitting a quadratic to points near the
        maximum.  Note that the data were collapsed to the left edge to
        get xdisp, so the location is the intercept on the edge, rather
        than where the spectrum crosses the middle of the detector.
    """

    y_nominal = xtract_info.field("b_spec")[0] + life_adj_offset
    segment = xtract_info.field("segment")[0]   # for possible warning message

    # The values of y_nominal and xd_range should be such that neither
    # y0 nor y1 will be less than zero or greater than 1023.
    y0 = int(round(y_nominal - xd_range))
    y1 = int(round(y_nominal + xd_range)) + 1
    if y0 < 0 or y1 >= len(xdisp):
        cosutil.printWarning("XD_RANGE in WCPTAB is too large.")
        y0 = max(y0, 0)
        y1 = min(y1, len(xdisp) - 1)

    boxcar_kernel = scipysignal.boxcar(box) / box
    xdisp_sm = ndimage.convolve(xdisp, boxcar_kernel, mode="nearest")
    len_xdisp_sm = len(xdisp_sm)

    if y0 >= y1:
        return (None, 0.)
    index = np.argsort(xdisp_sm[y0:y1])
    y = y0 + index[-1]
    signal = xdisp_sm[y]                # value in smoothed array
    # Check for duplicate values.
    y_min = y
    y_max = y
    while y_min > 0 and xdisp_sm[y_min] == signal:
        y_min -= 1
    while y_max < len_xdisp_sm and xdisp_sm[y_max] == signal:
        y_max += 1
    y_float = float(y_min + y_max) / 2.
    y = int(round(y_float))

    # Fit a quadratic to the smoothed curve near the peak.
    fit_range = (y_max - y_min) + box
    if fit_range < xd_range:
        r0 = y - fit_range // 2
        r1 = r0 + fit_range
        r0 = max(r0, 0)
        r1 = min(r1, len_xdisp_sm)
        r0 = r1 - fit_range
        x = np.arange(fit_range, dtype=np.float64)
        (coeff, var) = cosutil.fitQuadratic(x, xdisp_sm[r0:r1])
        (y_temp, y_float_sigma) = cosutil.centerOfQuadratic(coeff, var)
        if y_temp is None:
            return (None, 0.)
        y_float = y_temp + r0

    # Find the background level.
    i = index[(y1-y0)//2]
    background = xdisp_sm[y0+i]         # median of smoothed array

    sigma_s = math.sqrt(signal * box)
    sigma_b = math.sqrt(background * box)
    sigma_s_b = math.sqrt(sigma_s**2 + sigma_b**2)
    if sigma_s_b > 0.:
        signal_to_noise = (signal - background) * box / sigma_s_b
    else:
        signal_to_noise = 0.

    if signal_to_noise >= 5.:
        shift2 = y_float - y_nominal + life_adj_offset
    else:
        shift2 = None

    return (shift2, y_float)

def printWavecalRef(reffiles):
    """Print the names of reference files used for wavecal processing.

    Parameters
    ----------
    reffiles: dictionary
        Reference file names
    """

    cosutil.printRef("wcptab", reffiles)
    cosutil.printRef("lamptab", reffiles)
    cosutil.printRef("xtractab", reffiles)



"""Tests for COS/NUV G185M sci data."""

import pytest

import calcos
from helpers import BaseCOS


# TODO: Mark this as slow when there are faster tests added for CI tests
#       so that this only runs in nightly tests.
@pytest.mark.slow
class TestNUVSciG185M(BaseCOS):
    detector = 'nuv'

    def test_nuv_sci_g185m(self):
        """
        COS regression test
        """
        files_to_download = ['la8q99050_asn.fits',
                             'la8q99jbq_rawtag.fits',
                             'la8q99jbq_spt.fits']

        # Prepare input files.
        self.get_input_files(files_to_download)

        input_file = 'la8q99050_asn.fits'
        # Run CALCOS
        calcos.calcos(input_file)

        # Compare results.
        # The first outroot is the output from whole ASN,
        # the rest are individual members.
        outroots = ['la8q99050', 'la8q99jbq']
        outputs = []
        for sfx in ['x1dsum', 'x1dsum3']:
            fname = f'{outroots[0]}_{sfx}.fits'
            comparison_name = 'ref_' + fname
            outputs.append((fname, comparison_name))
        for outroot in outroots[1:]:
            for sfx in ('corrtag', 'counts', 
                        'flt', 'lampflash', 'x1d'):
                fname = '{}_{}.fits'.format(outroot, sfx)
                outputs.append((fname, 'ref_' + fname))
        self.compare_outputs(outputs, rtol=1e-7)



import math
import numpy as np
from .calcosparam import *
from . import cosutil
from . import ccos
LARGE_BURST = -20               # flag value in bkg_counts
SMALL_BURST = -10               # flag value in bkg_counts

def burstFilter(time, y, dq, reffiles, info, burstfile=None,
                high_countrate=False):
    """Flag regions where the count rate is unreasonably high.

    For each burst interval detected, a flag will be set in the
    data quality column for each event within that time interval.
    This is based on cf_screen_burst.c for FUSE.

    Parameters
    ----------
    time: array_like
        The time column in the events table.

    y: array_like
        The column of cross dispersion locations of events.

    dq: array_like
        The data quality column in the events table (updated in-place).

    reffiles: dictionary
        Reference file names.

    info: dictionary
        Header keywords and values.

    burstfile: str or None
        Name of output text file for burst info.

    high_countrate: boolean
        This flag can be set to true to force the time interval to be
        short, as if the data were high count rate.

    Returns
    -------
    bursts: list of two floats
        List of [bad_start, bad_stop] intervals during which a burst
        was detected (seconds since expstart).
    """

    bursts = None

    if info["segment"][:3] != "FUV":
        return None

    if info["exptime"] <= 0.:
        cosutil.printWarning("burstFilter:  Can't screen for bursts because"
                             " exptime = %g" % info["exptime"])
        return None

    cosutil.printMsg("Screen for bursts", VERBOSE)

    # Read parameters from burst reference table.
    (median_n, delta_t, delta_t_high, median_dt, burst_min,
        stdrej, source_frac, max_iter, high_rate) = \
                getBurstParam(reffiles["brsttab"], info["segment"])

    # Read location of active area from baseline reference table and
    # source and background locations from 1-D extraction reference table.
    try:
        (active_low, active_high, src_low, src_high,
         bkg1_low, bkg1_high, bkg2_low, bkg2_high) = \
                getRegionLocations(reffiles, info)
    except:
        cosutil.printWarning("Can't screen for bursts" \
                             " due to missing row in reference table")
        return None

    # rows of extraction aperture / background rows
    bkgsf = float(src_high - src_low + 1) / \
            float(bkg1_high - bkg1_low + bkg2_high - bkg2_low + 2)

    exptime = info["exptime"]
    countrate = float(len(time)) / exptime
    cosutil.printMsg("Total counts = %d, exposure time = %g s,"
            " count rate = %g c/s" % (len(time), exptime, countrate), VERBOSE)

    t0 = time[0]
    last = time[len(time)-1]
    if last <= t0:
        return None

    if countrate > high_rate:
        delta_t = delta_t_high
        cosutil.printMsg("High count rate; time bin set to %g s" % delta_t,
                         VERBOSE)
    elif high_countrate:
        delta_t = delta_t_high
        cosutil.printMsg("time bin set to %g s" % delta_t, VERBOSE)
    nbins = int(math.ceil((last - t0) / delta_t))
    del countrate, last
    if nbins <= 3:
        cosutil.printWarning("There are so few time bins (%d) that "
                "burst detection is not practical." % nbins)
        return None

    printParam(median_n, delta_t, median_dt, burst_min,
               stdrej, source_frac, max_iter, high_rate,
               active_low, active_high, src_low, src_high,
               bkg1_low, bkg1_high, bkg2_low, bkg2_high)

    # istart & istop are arrays of indices for slicing up the time and y
    # columns into intervals of length delta_t seconds.
    istart = np.zeros(nbins, dtype=np.int32)
    istop = np.zeros(nbins, dtype=np.int32)
    bkg_counts = np.zeros(nbins, dtype=np.int32)
    src_counts = np.zeros(nbins, dtype=np.int32)

    # Find istart & istop for each delta_t interval.
    ccos.getstartstop(time, istart, istop, delta_t)

    # find the counts within the background and source regions, for each
    # delta_t interval.
    ccos.getbkgcounts(y, dq, istart, istop,
                      bkg_counts, src_counts,
                      bkg1_low, bkg1_high, bkg2_low, bkg2_high,
                      src_low, src_high, bkgsf)

    bkg_counts_save = bkg_counts.copy()

    # Find the median of the values in the background counts array.
    index = np.argsort(bkg_counts)
    mid = nbins // 2
    median = bkg_counts[index[mid]]
    del index
    if median < 1:
        cosutil.printWarning("median = %d is unreasonable, reset to 1."
                             % median, VERBOSE)
        median = 1
    cutoff = median_n * median
    cosutil.printMsg("Initial check for large bursts,", VERBOSE)
    cosutil.printMsg("  median = %d, cutoff = %g; time interval = %.1f"
                     % (median, cutoff, delta_t), VERBOSE)
    # Identify intervals where the counts are greater than median_n * median.
    b1_flags = bkg_counts > cutoff
    index = np.nonzero(b1_flags)[0]
    nreject = len(index)
    for k in range(nreject):
        i = index[k]
        cosutil.printMsg("large burst at time %d, counts = %d"
                         % (int(time[istart[i]] + delta_t/2.), bkg_counts[i]),
                         VERBOSE)
        # Flag all events in the interval.
        dq[istart[i]:istop[i]] |= DQ_BURST
    if nreject > 0:
        # Set bkg_counts to a negative value for each burst.
        bkg_counts = np.where(b1_flags, LARGE_BURST, bkg_counts)
        cosutil.printMsg("%d large bursts detected." % nreject, VERBOSE)
    else:
        cosutil.printMsg("No large burst detected.", VERBOSE)
    del b1_flags

    # Search for smaller bursts.
    cosutil.printMsg("Check for smaller bursts;", VERBOSE)
    cosutil.printMsg("  median filter over time = %d s" % median_dt, VERBOSE)
    smallest_burst = burst_min * delta_t
    half_block = int(round(median_dt / delta_t)) // 2
    ccos.smallerbursts(time, dq,
                       istart, istop, bkg_counts, src_counts,
                       delta_t, smallest_burst, stdrej, source_frac,
                       half_block, max_iter,
                       LARGE_BURST, SMALL_BURST, DQ_BURST,
                       cosutil.checkVerbosity(VERBOSE))

    if burstfile is not None:
        # Write the time (middle of the interval), the background counts,
        # and whether the interval was flagged as a burst, large or small.
        fd = open(burstfile, "a")
        for i in range(nbins):
            t = t0 + (i+0.5) * delta_t
            fd.write("%.3f %d %d %d\n" %
                     (t, bkg_counts_save[i],
                         bkg_counts[i] == LARGE_BURST,
                         bkg_counts[i] == SMALL_BURST))
        fd.close()

    # Construct the list of start, stop intervals containing bursts.
    bursts = extractIntervals(time, istart, bkg_counts)

    return bursts

def getBurstParam(brsttab, segment):
    """Read parameters from burst reference table.

    Parameters
    ----------
    brsttab: str
        The name of the burst reference table.

    segment: str {"FUVA", "FUVB"}
        FUV segment name.

    Returns
    -------
    tuple
        The parameters read from the brsttab.
    """

    burst_info = cosutil.getTable(brsttab, filter={"segment": segment},
                                  exactly_one=True)

    median_n     = burst_info.field("median_n")[0]
    delta_t      = burst_info.field("delta_t")[0]
    delta_t_high = burst_info.field("delta_t_high")[0]
    median_dt    = burst_info.field("median_dt")[0]
    burst_min    = burst_info.field("burst_min")[0]
    stdrej       = burst_info.field("stdrej")[0]
    source_frac  = burst_info.field("source_frac")[0]
    max_iter     = burst_info.field("max_iter")[0]
    high_rate    = burst_info.field("high_rate")[0]

    return (median_n, delta_t, delta_t_high, median_dt, burst_min,
            stdrej, source_frac, max_iter, high_rate)

def getRegionLocations(reffiles, info):
    """Read region locations from reference tables.

    The lower and upper limits of the active area will be read from the
    baseline reference table.  The location and height of the source
    extraction region will be read from the 1-D extraction parameters
    table, and these will be used to define the source and background
    regions.

    Parameters
    ----------
    reffiles: dictionary
        Reference file names.

    info: dictionary
        Header keywords and values.

    Returns
    -------
    tuple
        A tuple with the lower and upper limits (in the cross-dispersion
        direction) of the active area, the source extraction region, and
        the two background regions.
    """

    (active_low, active_high, active_left, active_right) = \
            cosutil.activeArea(info["segment"], reffiles["brftab"])

    filter = {"segment": info["segment"],
              "opt_elem": info["opt_elem"],
              "cenwave": info["cenwave"],
              "aperture": info["aperture"]}
    xtract_info = cosutil.getTable(reffiles["xtractab"],
                  filter, exactly_one=True)
    b_spec = xtract_info.field("b_spec")[0]
    b_spec = int(round(b_spec))
    height = xtract_info.field("height")[0]

    src_low = b_spec - height // 2
    src_high = b_spec + height // 2

    bkg1_low = max(0, active_low)
    bkg2_high = min(1023, active_high)
    bkg1_high = src_low - height // 4
    bkg2_low = src_high + height // 4

    if info["tagflash"]:
        # Reset bkg2_low to a point above the wavecal spectrum.
        filter["aperture"] = "WCA"
        xtract_info = cosutil.getTable(reffiles["xtractab"],
                                       filter, exactly_one=True)
        b_spec = xtract_info.field("b_spec")[0]
        b_spec = int(round(b_spec))
        bkg2_low = b_spec + height * 3 // 4

    return (active_low, active_high, src_low, src_high,
            bkg1_low, bkg1_high, bkg2_low, bkg2_high)

def extractIntervals(time, istart, bkg_counts):
    """Construct list of bad time intervals.

    Parameters
    ----------
    time: array_like
        Time column from events table.

    istart: array_like
        Array of indices; time[istart[i]] is the time at the start of bin i.

    bkg_counts: array_like
        Negative values are used to flag bursts (otherwise, this is the
        array of background counts within each time bin).

    Returns
    -------
    list of two-element lists, or None
        List of [bad_start, bad_stop] intervals during which a burst was
        detected (seconds since expstart).  The function value will be
        None of no burst was detected.
    """

    if bkg_counts.min() >= 0:
        return None

    bursts = []
    nbins = len(bkg_counts)

    in_bad_interval = False
    for i in range(nbins):
        if bkg_counts[i] < 0 and not in_bad_interval:
            in_bad_interval = True
            t1 = time[istart[i]]        # time at start of current bin
        elif bkg_counts[i] >= 0 and in_bad_interval:
            in_bad_interval = False
            t2 = time[istart[i]]        # time at end of previous bin
            bursts.append([t1, t2])

    if in_bad_interval:
        bursts.append([t1, time[-1]])

    return bursts

def printParam(median_n, delta_t, median_dt, burst_min,
               stdrej, source_frac, max_iter, high_rate,
               active_low, active_high, src_low, src_high,
               bkg1_low, bkg1_high, bkg2_low, bkg2_high):
    """Print the parameters that will be used."""

    if not cosutil.checkVerbosity(VERY_VERBOSE):
        return

    cosutil.printMsg("The burst parameters are:", VERY_VERBOSE)

    cosutil.printMsg(
"reject counts higher than %.1f times the global median" % median_n,
                VERY_VERBOSE)

    cosutil.printMsg(
"%.1f counts/s is considered to be high count rate" % high_rate,
                VERY_VERBOSE)

    cosutil.printMsg(
"time interval for binning events = %.1f s" % delta_t, VERY_VERBOSE)

    cosutil.printMsg(
"time interval for median filter = %.1f s" % median_dt, VERY_VERBOSE)

    cosutil.printMsg(
"%.1f counts/s is minimum count rate that can be regarded as a burst" %
                burst_min, VERY_VERBOSE)

    cosutil.printMsg(
"reject counts greater than %.1f standard deviations" % stdrej, VERY_VERBOSE)

    cosutil.printMsg(
"burst must exceed %.3f of the source count rate before it is" % source_frac,
                VERY_VERBOSE)

    cosutil.printMsg("  considered to be significant", VERY_VERBOSE)

    cosutil.printMsg("maximum number of iterations = %d" % max_iter,
                VERY_VERBOSE)

    cosutil.printMsg(
"active area is rows %d to %d inclusive" % (active_low, active_high),
                VERY_VERBOSE)

    cosutil.printMsg(
"source extraction region is rows %d to %d inclusive" % (src_low, src_high),
                VERY_VERBOSE)

    cosutil.printMsg(
"background regions are %d to %d and %d to %d inclusive" %
               (bkg1_low, bkg1_high, bkg2_low, bkg2_high), VERY_VERBOSE)



class ShiftFile(object):
    """Read shift_file.

    The shift_file is a text file supplied by the user to specify either
    shift1 or shift2 (or both).  Blank lines and lines beginning with '#'
    will be ignored; otherwise, every line must have either five or six
    words:

        rootname, fpoffset, flash_number, segment/stripe, shift1, shift2

    All words given as strings are case insensitive (internally they will be
    converted to lower case).
    flash_number is one-indexed, to agree with the header keywords and with
    the information written to the trailer file.
    If shift1 is not to be specified, use the value "N/A".
    If shift2 is not to be specified, it may be given as "N/A" or simply left
    off (i.e. only five words on the line).
    Any or all of rootname, fpoffset, flash_number and segment/stripe may be
    given as "ANY", which is interpreted as a wildcard, i.e. it matches any
    rootname, fpoffset, etc.

    user_shifts = shiftfile.ShiftFile(shift_file, rootname, fpoffset)

    getShifts is a public method:
        ((shift1, shift2), nfound) = user_shifts.getShifts(key)
    key is a tuple of flash number (or "any") and segment/stripe name.

    Parameters
    ----------
    shift_file: str
        Name of text file supplied by user.

    rootname: str
        Rootname of the current exposure.

    fpoffset: int
        Fpoffset of the current exposure.
    """

    def __init__(self, shift_file, rootname, fpoffset):
        """Constructor."""

        # This is a dictionary of shifts for the current exposure, with
        # keys (flash_number, segment) and values (shift1, shift2).
        self.user_shift_dict = None

        fd = open(shift_file, "r")
        lines = fd.readlines()
        fd.close()

        user_shift_dict = {}
        for line in lines:
            line = line.strip()
            if not line:                        # ignore blank lines
                continue
            if line[0] == '#':                  # ignore comments
                continue
            words = line.split()
            if not words:                       # ignore blank lines
                continue
            nwords = len(words)
            if nwords < 5 or nwords > 6:
                raise RuntimeError("error reading this line of "
                                   "shift_file:  '%s'" % line)
            for i in range(nwords):
                words[i] = words[i].lower()
            # Select rows matching rootname and fpoffset.
            if words[0] != "any" and rootname != words[0]:
                continue
            if words[1] != "any" and fpoffset != int(words[1]):
                continue
            if words[2] == "any":
                flash_number = "any"
            else:
                flash_number = int(words[2])
            segment = words[3].lower()          # could be "any"
            key = (flash_number, segment)
            if words[4] != "n/a":
                shift1 = float(words[4])
            else:
                shift1 = None
            if nwords == 6 and words[5] != "n/a":
                shift2 = float(words[5])
            else:
                shift2 = None
            user_shift_dict[key] = (shift1, shift2)

        self.user_shift_dict = user_shift_dict

    def getShifts(self, key):
        """Return the shifts corresponding to key, if any.

        Parameters
        ----------
        key: tuple
            Flash number (one indexed) and segment; if flash number
            is "any" it matches any flash number (use "any" for auto/GO
            wavecals), and if segment is "any" it matches any segment or
            stripe (strings are case insensitive)

        Returns
        -------
        tuple
            ((shift1, shift2), nfound), where nfound is the number of
            elements--which should be either 0 or 1--that match key;
            either shift1 or shift2 may be None, and they will both be
            None if nfound is 0
        """

        (flash_number, segment) = key
        segment = segment.lower()
        if isinstance(flash_number, str):
            flash_number = flash_number.lower()

        nfound = 0
        shifts = (None, None)
        # sf_key is the flash number and segment read from the shift file
        for sf_key in self.user_shift_dict.keys():
            if sf_key[0] == "any" or flash_number == "any" or \
                                     flash_number == sf_key[0]:
                if sf_key[1] == "any" or segment == "any" or \
                                         segment == sf_key[1]:
                    shifts = self.user_shift_dict[sf_key]
                    nfound += 1

        return (shifts, nfound)



import copy
import math
import numpy as np
from astropy.io import fits
from astropy.stats import poisson_conf_interval
from . import cosutil
from .calcosparam import *       # parameter definitions

# Extract a slice of this height from the flat field in Spectrum.
XD_WIDTH = 15

def fpAvgSpec(input, output):
    """Average 1-D extracted FP-POS spectra.

    It is assumed that the arrays in all the input tables have the same
    length, but the output spectra will in general be longer than the
    input spectra.  The wavelengths in the input spectra will cover
    different ranges if the inputs are for different FP-POS positions.

    For FP-POS observations, the spectrum is moved over the detector
    from one exposure to another by tilting the grating.  This actually
    changes the dispersion relation, not just the zero point.  We assume,
    however, that the effect is the same as if we moved the detector with
    respect to a fixed spectrum.  That is, we use the same dispersion
    relation for all FP-POS positions, but the pixel numbers are offset
    by an amount that is constant for each FP-POS position.

    The data in the input table are count rates or fluxes.  When
    averaging the input spectra, we therefore weight by the exposure
    time.  For FUV data, the input spectra will be aligned to the
    nearest pixel and averaged without interpolation.  For NUV data,
    the pixel offset between input spectra need not be an integer.
    The output pixel size is the same as the input pixel size, and
    the contribution of a given input pixel to an output pixel is
    proportional to the area of overlap.  This is equivalent to linear
    interpolation, since the area of overlap (actually the length, since
    this is 1-D) is linearly related to the pixel shift.

    Parameters
    ----------
    input: list of str
        Name(s) of the input x1d files.

    output: str
        Name of a file for the averaged spectra.
    """

    nfiles = len(input)

    assert nfiles >= 1

    cosutil.printIntro("Average 1-D spectra")
    names = [("Input", repr(input)), ("Output", output)]
    cosutil.printFilenames(names)

    if nfiles == 1:
        oneInputFile(input[0], output)
    else:
        outspec = OutputX1D(input, output)

def oneInputFile(input, output):
    """Copy input to output, setting values to zero if dq_wgt is zero.

    Parameters
    ----------
    input: str
        Name of the (one) input x1d file

    output: str
        Name of a file for the modified copy of input
    """

    fd = fits.open(input, mode="copyonwrite")
    data = fd[1].data
    if data is None or len(data) == 0:
        fd.close()
        cosutil.copyFile(input, output)
        return

    flux = data.field("flux")
    error = data.field("error")
    gross = data.field("gross")
    error_lower = data.field("error_lower")
    err_frequentist_up = data.field("error")
    net = data.field("net")
    background = data.field("background")
    dq_wgt = data.field("dq_wgt")
    gcounts = data.field("gcounts")
    error = data.field("error")
    variance_flat = data.field("variance_flat")
    variance_counts = data.field("variance_counts")
    variance_bkg = data.field("variance_bkg")

    for row in range(len(data)):
        flux[row,:] = np.where(dq_wgt[row] <= 0., 0., flux[row])
        error[row,:] = np.where(dq_wgt[row] <= 0., 0., error[row])
        gross[row,:] = np.where(dq_wgt[row] <= 0., 0., gross[row])
        gcounts[row,:] = np.where(dq_wgt[row] <= 0., 0., gcounts[row])
        error_lower[row,:] = np.where(dq_wgt[row] <= 0., 0., error_lower[row])
        net[row,:] = np.where(dq_wgt[row] <= 0., 0., net[row])
        background[row,:] = np.where(dq_wgt[row] <= 0., 0., background[row])
        variance_flat[row,:] = np.where(dq_wgt[row] <= 0., 0., variance_flat[row])
        variance_counts[row,:] = np.where(dq_wgt[row] <= 0., 0., variance_counts[row])
        variance_bkg[row,:] = np.where(dq_wgt[row] <= 0., 0., variance_bkg[row])

    cosutil.updateFilename(fd[0].header, output)
    if cosutil.isProduct(output):
        asn_mtyp = fd[1].header.get("asn_mtyp", "missing")
        asn_mtyp = cosutil.modifyAsnMtyp(asn_mtyp)
        if asn_mtyp != "missing":
            fd[1].header["asn_mtyp"] = asn_mtyp
    list_keywords = [("MFPPOS", "fppos", fd[0].header["fppos"]),
                     ("MFPOFSET", "fpoffset", fd[0].header["fpoffset"]),
                     ("MCENWAVE", "cenwave", fd[0].header["cenwave"])]
    for (new_kwd, old_kwd, value) in list_keywords:
        str_value = str(value)
        fd[0].header.set(new_kwd, str_value, after=old_kwd)     # xxx temp
        # xxx fd[0].header.insert(old_kwd, (new_kwd, str_value), after=True)
    del_these = ["segment", "wavecals", "fppos", "fpoffset"]
    for keyword in del_these:
        if keyword in fd[0].header:
            del(fd[0].header[keyword])
    delSomeKeywords(fd[1].header)
    newfd = delExtraColumns(fd[1])
    fd[1] = newfd
    fd.writeto(output)
    fd.close()

def delSomeKeywords(hdr):
    """Delete exposure-specific keywords.

    Parameters
    ----------
    hdr: ``astropy.io.fits.header.Header`` object
        Extension header to be modified
    """

    # These keywords are exposure-specific and are not relevant
    # to the entire association.
    for key in ["shift1a", "shift1b", "shift1c",
                "shift2a", "shift2b", "shift2c",
                "dpixel1a", "dpixel1b", "dpixel1c"]:
        if key in hdr:
            del(hdr[key])

def delExtraColumns(hdu):
    """Delete extra columns in the output file.

    Parameters
    ----------
    hdu: ``astropy.io.fits.hdu.table.BinTableHDU`` object
        HDU object to be modified to be modified
    """

    columns_to_delete = ["DQ_OUTER", "BACKGROUND_PER_ROW", "NUM_EXTRACT_ROWS",
                         "ACTUAL_EE", "Y_LOWER_OUTER", "Y_UPPER_OUTER",
                         "Y_LOWER_INNER", "Y_UPPER_INNER"]

    table = hdu.data
    header = hdu.header
    columns = table.columns
    outcols = []
    for column in columns:
        if column.name in columns_to_delete:
            number = 0
            while(True):
                number = number + 1
                tstring = 'TTYPE' + str(number)
                try:
                    ttype = header[tstring]
                    if ttype in columns_to_delete:
                        del header[tstring]
                        tstring = 'TFORM' + str(number)
                        del header[tstring]
                        tstring = 'TUNIT' + str(number)
                        try:
                            del header[tstring]
                        except KeyError:
                            pass
                        tstring = 'TDISP' + str(number)
                        try:
                            del header[tstring]
                        except KeyError:
                            pass
                except KeyError:
                    break
        else:
            outcols.append(column)
    cd = fits.ColDefs(outcols)
    newhdu = fits.BinTableHDU.from_columns(cd, header=hdu.header)
    return newhdu

def makeStringList(inlist):
    """Construct a comma-separated string from elements of a list.

    Parameters
    ----------
    inlist: list
        Items to be written to a string.

    Returns
    -------
    outstr: str
        The elements of inlist as a string, with a comma and blank
        between each element.
    """

    nelem = len(inlist)
    outstr = ""
    for i in range(nelem):
        if i == 0:
            outstr += "%s" % str(inlist[i])
        else:
            outstr += ", %s" % str(inlist[i])

    return outstr

def pixelsFromWl(input_wavelength, output_wavelength):
    """Find pixel numbers in input corresponding to wavelengths in output.

    This function returns an array of pixel coordinates (floating point)
    in the input spectrum that have the same wavelengths as pixels 0, 1, 2,
    etc., in the output spectrum.

    An example may help.  Suppose pixel k in the output spectrum has
    wavelength wl (i.e. output_wavelength[k] = wl).  Suppose pixel n (an
    integer, just for example) in the input spectrum has the same wavelength
    wl.  Then pixel k in the array returned by this function would have
    value n.

    input_wavelength and output_wavelength do not need to have the same
    length.  As used by fpAvgSpec, there will typically be wavelengths in
    output_wavelength that lie outside the range of input_wavelength.  For
    those points, the element in the returned array can be less than or equal
    to zero, or it could be greater than nelem-1; these values should not be
    counted on to be accurate extrapolations.

    Parameters
    ----------
    input_wavelength: array_like
        Array of wavelengths in input spectrum

    output_wavelength: array_like
        Array of wavelengths in output spectrum

    Returns
    -------
    array_like, same data type as wavelength arrays
        Pixel numbers (but not integer values) in input spectrum
    """

    nelem = len(input_wavelength)

    avgdisp = (input_wavelength[-1] - input_wavelength[0]) / (nelem - 1.)

    # disp will be the dispersion at each pixel of the input wavelengths.
    disp = input_wavelength.copy()
    disp[1:nelem-1] = (input_wavelength[2:nelem] -
                       input_wavelength[0:nelem-2]) / 2.
    disp[0] = input_wavelength[1] - input_wavelength[0]
    disp[nelem-1] = input_wavelength[nelem-1] - input_wavelength[nelem-2]

    # x0 is a first estimate of the pixel numbers.
    x0 = (output_wavelength - input_wavelength[0]) / avgdisp
    x0 = np.where(x0 < 0., 0., x0)
    ix0 = x0.astype(np.int32)
    ix0 = np.where(ix0 > nelem-1, nelem-1, ix0)

    # wavelengths in input at pixels ix0 are input_wavelength[ix0]
    diff = (output_wavelength - input_wavelength[ix0])

    # x1 should be very close to the correct pixel numbers.
    x1 = ix0 + diff / disp[ix0]
    x1 = np.where(x1 < 0., 0., x1)
    ix1 = x1.astype(np.int32)
    ix1 = np.where(ix1 > nelem-1, nelem-1, ix1)

    diff = (output_wavelength - input_wavelength[ix1])
    ipixel = ix1 + diff / disp[ix1]

    return ipixel

class OutputX1D(object):
    """Average 1-D FP-POS spectra.

    Parameters
    ----------
    input: list of str
        Input file names

    output: str
        Output file name

    Attributes
    ----------
    input
    output

    keywords: dictionary
        Relevant keywords and values, e.g. detector

    inspec: list of Spectrum objects
        Input spectra

    segments: list of str
        Segment names found in input x1d tables

    ofd: ``astropy.io.fits.hdu.hdulist.HDUList`` object
        Astropy FITS object for output file

    nrows: int
        Number of rows to be written to the output table

    output_nelem: int
        Number of elements to use when allocating output arrays

    output_wl_range: dictionary of two-element tuples
        Smallest and largest wavelengths in an output spectrum; key is
        segment or stripe

    output_dispersion: dictionary
        Dispersion (Angstroms per pixel) in an output spectrum, key is
        segment or stripe
    """

    def __init__(self, input, output):
        """Constructor."""

        self.input = input
        self.output = output
        self.keywords = {}
        self.inspec = []
        self.segments = []
        self.ofd = None
        self.nrows = 0
        self.output_nelem = 1
        self.output_wl_range = {}
        self.output_dispersion = {}
        # This is the index of the element of self.inspec that has the
        # maximum value of nelem.  We'll use this spectrum as the template
        # for column definitions for the output table.
        self.index_max_nelem = 0

        # Create a list of Spectrum objects, and get info from headers.
        self.getInputInfo()

        # Check that the data in each Spectrum is comparable to the others.
        self.compareX1d()

        # Compute length of output arrays.
        self.computeOutputInfo()

        # Create ofd, the output HDUList object.
        self.createOutput()

        # Fill in the data in the output table.
        for segment in self.segments:           # for each output row ...
            osp = OutputSpectrum(self.ofd, self.inspec, self.keywords,
                                 segment, self.output_wl_range[segment],
                                 self.output_dispersion[segment])
        if cosutil.isProduct(self.output):
            asn_mtyp = self.ofd[1].header.get("asn_mtyp", "missing")
            asn_mtyp = cosutil.modifyAsnMtyp(asn_mtyp)
            if asn_mtyp != "missing":
                self.ofd[1].header["asn_mtyp"] = asn_mtyp
        self.updateArchiveSearch(self.ofd)      # minwave & maxwave
        newhdu = delExtraColumns(self.ofd[1])
        self.ofd[1] = newhdu
        self.ofd.writeto(self.output)

        if self.keywords["statflag"]:
            cosutil.doSpecStat(self.output)

    def getInputInfo(self):
        """Get info and data from input files.

        This routine creates Spectrum objects (one for each row of each input
        table) and appends them to the inspec list, gets keywords from the
        input headers, and determines the number of rows (nrows) that the
        output table should have.
        """

        # These are for averaging the global count rates.  The elements are
        # for NUV, FUVA and FUVB respectively.
        sum_globrate = [0., 0., 0.]     # incremented for each row in each file
        sum_exptime = [0., 0., 0.]      # exptime from the column
        avg_globrate = [-1., -1., -1.]  # average values for NUV, FUVA, FUVB
        # This is for updating the exptime, exptimea, exptimeb keywords.
        sum_exptime_kwd = [0., 0., 0.]  # exptime from the header keywords

        first = True            # true for first input file
        for input in self.input:
            ifd = fits.open(input, mode="readonly")
            phdr = ifd[0].header
            hdr = ifd[1].header
            # Get keyword values.
            if first:
                detector = phdr["detector"]
                opt_elem = phdr["opt_elem"]
                cenwave_list = [phdr["cenwave"]]    # may have multiple values
                fppos_list = [phdr["fppos"]]        # may have multiple values
                fpoffset_list = [phdr["fpoffset"]]  # may have multiple values
                (aperture, message) = cosutil.getApertureKeyword(phdr)
                statflag = phdr.get("statflag", False)
                sum_plantime = hdr["plantime"]
                expstart = hdr["expstart"]
                expend = hdr["expend"]
                first = False
            else:
                sum_plantime += hdr["plantime"]
                expstart = min(expstart, hdr["expstart"])
                expend = max(expend, hdr["expend"])
                if phdr["cenwave"] not in cenwave_list:
                    cenwave_list.append(phdr["cenwave"])
                if phdr["fppos"] not in fppos_list:
                    fppos_list.append(phdr["fppos"])
                if phdr["fpoffset"] not in fpoffset_list:
                    fpoffset_list.append(phdr["fpoffset"])
            fpoffset = phdr["fpoffset"]

            if ifd[1].data is not None:
                nrows = len(ifd[1].data)
                # for each row in the current input table
                for row in range(nrows):
                    sp = Spectrum(ifd, row, fpoffset)
                    segment = sp.segment
                    if segment not in self.segments:
                        self.segments.append(segment)
                    self.inspec.append(sp)
                    if sp.segment == "FUVA":
                        sum_exptime_kwd[1] += hdr.get("exptimea",
                                              default=hdr.get("exptime", 0.))
                        globrate = hdr.get("globrt_a", -1.)
                        if globrate >= 0.:
                            sum_globrate[1] += (globrate * sp.exptime)
                            sum_exptime[1] += sp.exptime
                    elif sp.segment == "FUVB":
                        sum_exptime_kwd[2] += hdr.get("exptimeb",
                                              default=hdr.get("exptime", 0.))
                        globrate = hdr.get("globrt_b", -1.)
                        if globrate >= 0.:
                            sum_globrate[2] += (globrate * sp.exptime)
                            sum_exptime[2] += sp.exptime
                    elif detector == "NUV" and row == 0:
                        sum_exptime_kwd[0] += hdr["exptime"]
                        globrate = hdr["globrate"]
                        if globrate >= 0.:
                            sum_globrate[0] += (globrate * sp.exptime)
                            sum_exptime[0] += sp.exptime

            ifd.close()

        for i in range(3):
            if sum_exptime[i] > 0.:
                avg_globrate[i] = sum_globrate[i] / sum_exptime[i]

        # number of rows to be written to the output table
        self.nrows = len(self.segments)

        if detector == "NUV":
            exptime = sum_exptime_kwd[0]
        else:
            exptime = max(sum_exptime_kwd[1], sum_exptime_kwd[2])
        self.keywords = {
             "detector": detector,
             "opt_elem": opt_elem,
             "cenwave":  cenwave_list,          # this is a list
             "fppos":    fppos_list,            # this is a list
             "fpoffset": fpoffset_list,         # this is a list
             "aperture": aperture,
             "exptime":  exptime,
             "exptimea": sum_exptime_kwd[1],
             "exptimeb": sum_exptime_kwd[2],
             "expstart": expstart,
             "expend":   expend,
             "expstrtj": expstart + MJD_TO_JD,
             "expendj":  expend + MJD_TO_JD,
             "plantime": sum_plantime,
             "globrate": avg_globrate[0],       # average for NUV spectra
             "globrt_a": avg_globrate[1],       # average for FUVA spectra
             "globrt_b": avg_globrate[2],       # average for FUVB spectra
             "statflag": statflag}

    def compareX1d(self):
        """Check that the rows of two x1d tables contain comparable info.

        Currently, the only check is on the array sizes.
        """

        for sp in self.inspec:
            if sp.nelem != self.inspec[0].nelem:
                raise RuntimeError("x1d tables have different array sizes.")

    def computeOutputInfo(self):
        """Compute length of output arrays, and info for output wavelengths.

        This routine assigns values to the attributes output_nelem,
        index_max_nelem, output_wl_range, and output_dispersion.
        """

        if len(self.inspec) < 1:
            if self.keywords["detector"] == "FUV":
                self.output_nelem = FUV_EXTENDED_X
            elif self.keywords["detector"] == "NUV":
                self.output_nelem = NUV_EXTENDED_X
            else:
                self.output_nelem = 1
            return

        # Find the maximum input nelem, and set output_nelem to that value.
        # The input nelem should really be all the same, but just in case
        # one of them is zero, we need to be able to skip that one.
        # Also set self.index_max_nelem, which we'll use in createOutput
        # for getting the initial column definitions for the output table.
        min_output_nelem = -1
        for (k, sp) in enumerate(self.inspec):
            if sp.nelem > min_output_nelem:
                self.index_max_nelem = k
                min_output_nelem = sp.nelem

        # Find the wavelength and dispersion for each segment.
        self.output_wl_range = {}
        self.output_dispersion = {}
        for segment in self.segments:
            min_wl = 1.e9
            max_wl = 1.
            min_dispersion = 1.e9
            for sp in self.inspec:
                if sp.segment == segment:
                    if sp.nelem < 2:
                        continue
                    min_wl = min(min_wl, sp.wavelength[0])
                    max_wl = max(max_wl, sp.wavelength[-1])
                    min_dispersion = min(min_dispersion,
                        (sp.wavelength[-1] - sp.wavelength[0]) / (sp.nelem - 1))
            self.output_wl_range[segment] = (min_wl, max_wl)
            self.output_dispersion[segment] = min_dispersion

        # Determine the number of elements we will need for the output
        # spectra.  The output array size should be at least as large as the
        # arrays in the input spectra (min_output_nelem)
        output_nelem = min_output_nelem         # initial value
        for segment in self.segments:
            (min_wl, max_wl) = self.output_wl_range[segment]
            dispersion = self.output_dispersion[segment]
            if dispersion <= 0.:
                continue
            nelem = (max_wl - min_wl) / dispersion
            nelem = int(round(math.ceil(nelem)))
            output_nelem = max(output_nelem, nelem)
        if output_nelem > 0:
            self.output_nelem = output_nelem

    def createOutput(self):
        """Create HDUList object for output file."""

        # Get header info from the input.
        ifd = fits.open(self.input[self.index_max_nelem],
                        mode="copyonwrite")
        detector = ifd[0].header["detector"]

        primary_hdu = fits.PrimaryHDU(header=ifd[0].header)
        cosutil.updateFilename(primary_hdu.header, self.output)
        # Add new keywords with comma-separated values.
        list_keywords = [("MFPPOS", "fppos", self.keywords["fppos"]),
                         ("MFPOFSET", "fpoffset", self.keywords["fpoffset"]),
                         ("MCENWAVE", "cenwave", self.keywords["cenwave"])]
        for (new_kwd, old_kwd, list_value) in list_keywords:
            str_value = makeStringList(list_value)
            primary_hdu.header.set(new_kwd, str_value, after=old_kwd)   # xxx
            # xxx primary_hdu.header.insert(old_kwd, (new_kwd, str_value),
            # xxx                           after=True)
        del_these = ["segment", "wavecals", "fppos", "fpoffset"]
        for keyword in del_these:
            if keyword in primary_hdu.header:
                del(primary_hdu.header[keyword])
        if len(self.keywords["cenwave"]) > 1:
            del(primary_hdu.header["cenwave"])
        ofd = fits.HDUList(primary_hdu)

        rpt = str(self.output_nelem)    # used for defining columns

        # Define the columns explicitly, rather than using an input table
        # as a template and then modifying the lengths of arrays (see below),
        # because the modified columns kept reverting to the original length.
        col = []
        col.append(fits.Column(name="SEGMENT", format="4A"))
        col.append(fits.Column(name="EXPTIME", format="1D",
                   disp="F8.3", unit="s"))
        col.append(fits.Column(name="NELEM", format="1J", disp="I6"))
        col.append(fits.Column(name="WAVELENGTH", format=rpt+"D",
                   unit="angstrom"))
        col.append(fits.Column(name="FLUX", format=rpt+"E",
                   unit="erg /s /cm**2 /angstrom"))
        col.append(fits.Column(name="ERROR", format=rpt+"E",
                   unit="erg /s /cm**2 /angstrom"))
        col.append(fits.Column(name="ERROR_LOWER", format=rpt+"E",
                   unit="erg /s /cm**2 /angstrom"))
        col.append(fits.Column(name="GROSS", format=rpt+"E",
                   unit="count /s"))
        col.append(fits.Column(name="GCOUNTS", format=rpt+"E",
                   unit="count"))
        col.append(fits.Column(name="VARIANCE_FLAT", format=rpt+"E",
                   unit="count"))
        col.append(fits.Column(name="VARIANCE_COUNTS", format=rpt+"E",
                   unit="count"))
        col.append(fits.Column(name="VARIANCE_BKG", format=rpt+"E",
                   unit="count"))
        col.append(fits.Column(name="NET", format=rpt+"E",
                   unit="count /s"))
        col.append(fits.Column(name="BACKGROUND", format=rpt+"E",
                   unit="count /s"))
        col.append(fits.Column(name="DQ", format=rpt+"I"))
        col.append(fits.Column(name="DQ_WGT", format=rpt+"E"))
        cd = fits.ColDefs(col)

        # Modify some of the output columns.
        #cd = ifd[1].columns             # this is a ColDefs object
        #col_names = cd.names
        #col_formats = cd.formats
        #ncols = len(col_names)
        # xxx x = ifd[1].data                 # xxx touch the data
        #for i in range(ncols):
        #    fmt = col_formats[i]
        #    if fmt[-1] in ["D", "E", "I", "J"] and fmt[0] in "123456789":
        #        x = ifd[1].data         # xxx touch the data
        #        newfmt = rpt + fmt[-1]
        #        cd.change_attrib(col_names[i], "format", newfmt)

        # Create output HDU for the table.
        newhdu = delExtraColumns(ifd[1])
        ifd[1].header = newhdu.header
        hdu = fits.BinTableHDU.from_columns(cd, header=ifd[1].header,
                                            nrows=self.nrows)

        hdu.header["exptime"] = self.keywords["exptime"]
        if detector == "FUV":
            hdu.header["exptimea"] = self.keywords["exptimea"]
            hdu.header["exptimeb"] = self.keywords["exptimeb"]

        hdu.header["expstart"] = self.keywords["expstart"]
        hdu.header["expend"] = self.keywords["expend"]
        hdu.header["expstrtj"] = self.keywords["expstrtj"]
        hdu.header["expendj"] = self.keywords["expendj"]
        hdu.header["plantime"] = self.keywords["plantime"]
        if self.keywords["globrate"] >= 0.:
            hdu.header["globrate"] = round(self.keywords["globrate"], 4)
        if self.keywords["globrt_a"] >= 0.:
            hdu.header["globrt_a"] = round(self.keywords["globrt_a"], 4)
        if self.keywords["globrt_b"] >= 0.:
            hdu.header["globrt_b"] = round(self.keywords["globrt_b"], 4)

        # Delete some keywords because they are specific to one exposure.
        delSomeKeywords(hdu.header)

        ofd.append(hdu)
        self.fpInitData(ofd)            # initialize data in output hdu

        ifd.close()

        self.ofd = ofd

    def fpInitData(self, ofd):
        """Initialize the output data block.

        Two scalar columns, SEGMENT and NELEM, will be set to their actual
        values.  EXPTIME and the array columns will be initialized to zero.
        """

        ofd[1].data.field("segment")[:] = self.segments
        ofd[1].data.field("nelem")[:] = self.output_nelem

        ofd[1].data.field("exptime")[:] = 0.

        ofd[1].data.field("wavelength")[:] = 0.
        ofd[1].data.field("flux")[:] = 0.
        ofd[1].data.field("error")[:] = 0.
        ofd[1].data.field("gross")[:] = 0.
        ofd[1].data.field("error_lower")[:] = 0.
        ofd[1].data.field("VARIANCE_FLAT")[:] = 0.
        ofd[1].data.field("VARIANCE_COUNTS")[:] = 0.
        ofd[1].data.field("VARIANCE_BKG")[:] = 0.
        ofd[1].data.field("gcounts")[:] = 0.
        ofd[1].data.field("net")[:] = 0.
        ofd[1].data.field("background")[:] = 0.
        ofd[1].data.field("dq")[:] = DQ_PIXEL_OUT_OF_BOUNDS
        ofd[1].data.field("dq_wgt")[:] = 0.

    def updateArchiveSearch(self, ofd):
        """Update the keywords giving min & max wavelengths.

        ofd: ``astropy.io.fits.hdu.hdulist.HDUList`` object
            For the output file, primary header modified in-place
        """

        phdr = ofd[0].header
        outdata = ofd[1].data
        nrows = outdata.shape[0]
        wavelength = outdata.field("WAVELENGTH")
        dq_wgt = outdata.field("DQ_WGT")

        if nrows <= 0 or len(wavelength[0]) < 1:
            return

        nelem = len(wavelength[0])
        # This initial value assumes wavelengths increase with pixel number.
        minwave = wavelength[0][nelem-1]
        maxwave = wavelength[0][0]
        for row in range(nrows):
            if dq_wgt[row].sum(dtype=np.float64) <= 0:
                good_wl = wavelength[row]
            else:
                good_wl = wavelength[row][dq_wgt[row] > 0.]
            minwave_row = good_wl.min()
            minwave = min(minwave, minwave_row)
            maxwave_row = good_wl.max()
            maxwave = max(maxwave, maxwave_row)

        phdr["MINWAVE"] = minwave
        phdr["MAXWAVE"] = maxwave
        phdr["BANDWID"] = maxwave - minwave
        phdr["CENTRWV"] = (maxwave + minwave) / 2.

class Spectrum(object):
    """One row of an input spectrum.

    Parameters
    ----------
    ifd: ``astropy.io.fits.hdu.hdulist.HDUList`` object
        The list of header/data objects for an input file

    row: int
        Row number (zero indexed) in the current input file

    fpoffset: int
        Value of the FPOFFSET keyword for the current input file

    Attributes
    ----------
    exptime: float
        exposure time (seconds) for this input spectrum

    segment: str
        segment or stripe name for the current row

    nelem: int
        number of elements in the arrays

    wavelength: array_like
        wavelengths for the current row

    flux: array_like
        flux values

    error: array_like
        error estimates for the flux

    gross: array_like
        gross values (count rate)

    gcounts: array_like
        gross counts

    net: array_like
        net values

    background: array_like
        background values

    dq: array_like
        data quality flags

    dq_wgt: array_like
        weights to account for pixels excluded due to data quality

    fpoffset
    """

    def __init__(self, ifd, row=0, fpoffset=0):
        """Constructor."""

        self.segment = ifd[1].data.field("segment")[row]
        self.exptime = ifd[1].data.field("exptime")[row]
        self.nelem = ifd[1].data.field("nelem")[row]
        self.wavelength = ifd[1].data.field("wavelength")[row]
        self.flux = ifd[1].data.field("flux")[row]
        self.error = ifd[1].data.field("error")[row]
        self.gross = ifd[1].data.field("gross")[row]
        self.error_lower = ifd[1].data.field("error_lower")[row]
        self.variance_flat = ifd[1].data.field("variance_flat")[row]
        self.variance_counts = ifd[1].data.field("variance_counts")[row]
        self.variance_bkg = ifd[1].data.field("variance_bkg")[row]
        self.gcounts = ifd[1].data.field("gcounts")[row]
        self.net = ifd[1].data.field("net")[row]
        self.background = ifd[1].data.field("background")[row]
        self.dq = ifd[1].data.field("dq")[row]
        self.dq_wgt = ifd[1].data.field("dq_wgt")[row]
        self.fpoffset = fpoffset

        # These are used for weighting by the flat field.
        self.origin_ff = np.zeros(2, dtype=np.int32)
        self.data_ff = None
        self.state_ff = "empty"

        # Read the flat field.
        phdr = ifd[0].header
        obsmode = phdr["obsmode"]
        doppcorr = phdr["doppcorr"]
        detector = phdr["detector"]
        hdr = ifd[1].header
        seg = self.segment[-1].lower()
        if phdr.get("exptype", default="missing").startswith("EXTERNAL"):
            shift1 = hdr["shift1" + seg]
            shift2 = hdr["shift2" + seg]
        else:
            shift1 = 0.
            shift2 = 0.
        sp_loc = hdr["sp_loc_" + seg]
        x_offset = hdr.get("x_offset", default=0)
        expstart = hdr["expstart"]
        expend = hdr["expend"]

        self.getFlatField(ifd)
        if self.data_ff is not None:
            self.collapseFlatField(sp_loc, shift2, width=XD_WIDTH)
            (doppmag, doppzero, orbitper) = \
                        self.getDopplerParam(doppcorr, obsmode,
                                             detector, hdr)
            self.shiftFlatField(shift1, x_offset,
                                doppcorr, doppmag, doppzero, orbitper,
                                expstart, expend)

    def getFlatField(self, ifd):
        """Read the flat field data.

        Values will be assigned to self.origin_ff, self.data_ff, and
        self.state_ff.

        Parameters
        ----------
        ifd: ``astropy.io.fits.hdu.hdulist.HDUList`` object
            The list of header/data objects for the current input file
        """

        phdr = ifd[0].header
        flatcorr = phdr["flatcorr"]
        if flatcorr != "COMPLETE" and flatcorr != "PERFORM":
            return
        flatfile = phdr["flatfile"]
        detector = phdr["detector"]
        segment = self.segment
        reffiles = {}
        reffiles["flatfile"] = cosutil.expandFileName(flatfile)
        fd = fits.open(reffiles["flatfile"])
        if detector == "FUV":
            self.data_ff = fd[(segment,1)].data.copy()
            hdr_ff = fd[(segment,1)].header
        else:
            self.data_ff = fd[1].data.copy()
            hdr_ff = fd[1].header
        self.origin_ff[0] = hdr_ff.get("origin_y", default=0)
        self.origin_ff[1] = hdr_ff.get("origin_x", default=0)
        fd.close()
        self.state_ff = "2-D, as read from reference file"

    def collapseFlatField(self, sp_loc, shift2, width):
        """Average a slice of self.data_ff, replacing full array.

        Parameters
        ----------
        sp_loc: float
            The value of keyword SP_LOC_[ABC], the Y location where the
            spectrum was extracted (where it crossed the middle of the
            detector).

        shift2: float
            The wavecal offset in Y.

        width: float
            The width in Y (i.e. the height) of the slice of the flat
            field to be extracted.
        """

        if self.data_ff is None:
            return
        height_ff = self.data_ff.shape[0]
        if height_ff < 1:
            return

        lowlim = sp_loc - width // 2 + shift2 - self.origin_ff[0]
        lowlim = int(round(lowlim))
        highlim = int(lowlim + width)        # upper limit of a slice
        if highlim < 0 or lowlim >= height_ff:
            cosutil.printWarning("Target is outside range of flat field")
            self.data_ff = self.data_ff.mean(axis=0, dtype=np.float64)
        else:
            lowlim = max(lowlim, 0)
            highlim = min(highlim, height_ff)
            height = highlim - lowlim           # height of slice
            if height > 0:
                self.data_ff = \
                self.data_ff[lowlim:highlim,:].mean(axis=0,
                                                  dtype=np.float64)
            else:
                cosutil.printError("Height of flat field slice is %d" % height)
                self.data_ff = self.data_ff.mean(axis=0, dtype=np.float64)
        self.state_ff = "slice taken, averaged to 1-D"

    def getDopplerParam(self, doppcorr, obsmode, detector, hdr):
        """Read Doppler parameters from the header.

        Parameters
        ----------
        doppcorr: str
            PERFORM, COMPLETE, OMIT, SKIPPED.

        obsmode: str
            TIME-TAG or ACCUM.

        detector: str
            FUV or NUV.

        hdr: ``astropy.io.fits.header.Header`` object
            The header of the first extension.

        Returns
        -------
        (doppmag, doppzero, orbitper): tuple of three floats
        """

        if doppcorr == "OMIT" or doppcorr == "SKIPPED":
            doppmag  = 0.
            doppzero = 0.
            orbitper = 5760.
        elif obsmode == "TIME-TAG":
            # Get wavelength and dispersion from spectrum near middle.
            nelem = len(self.wavelength)
            if detector == "FUV":
                middle = nelem // 2
            else:
                middle = hdr.get("x_offset", default=0) + NUV_X // 2
            wavelength = self.wavelength[int(middle)]
            low = int(middle - 50)
            high = int(middle + 50)
            if low > 0 and high < nelem:
                disp = (self.wavelength[high] - self.wavelength[low]) / 100.
            else:
                raise RuntimeError("wavelength array is too short, %d" % nelem)
            doppmag = (hdr["doppmagv"] / SPEED_OF_LIGHT) * (wavelength / disp)
            doppzero = hdr["doppzero"]
            orbitper = hdr["orbitper"]

        else:               # ACCUM
            doppmag  = hdr["dopmagt"]
            doppzero = hdr["dopzerot"]
            orbitper = hdr["orbtpert"]

        return (doppmag, doppzero, orbitper)

    def shiftFlatField(self, shift1, x_offset,
                       doppcorr, doppmag, doppzero, orbitper,
                       expstart, expend):
        """

        Parameters
        ----------
        shift1: float
            OSM (wavecal) shift in dispersion direction.

        x_offset: int
            Value of X_OFFSET keyword (buffer to avoid losing data due to
            fpoffset shift).

        doppcorr: str
            PERFORM, OMIT, etc.

        doppmag: float
            Magnitude of Doppler shift in pixels.

        doppzero: float
            Time (MJD) when the orbital Doppler shift is zero and
            increasing.

        orbitper: float
            Orbital period (seconds) of HST.

        expstart: float
            Exposure start time (MJD).

        expend: float
            Exposure end time (MJD).
        """

        if self.data_ff is None:
            return
        nelem = len(self.wavelength)
        if nelem < 1:
            return
        if len(self.data_ff) > nelem:
            cosutil.printWarning("Flat field is wider than data array,"
                                 " %d vs %d;" % (len(self.data_ff), nelem))
            cosutil.printContinuation("flat will not be included in weight.")
            self.data_ff = None
            return

        flat = np.ones(nelem, dtype=np.float64)

        # dispersion direction
        shift1 = round(shift1)
        offset1 = self.origin_ff[1] - shift1 + x_offset
        offset1 = int(round(offset1))
        len_ff = len(self.data_ff)
        # Check for extending off either end of flat.
        if offset1 < 0 or offset1 + len_ff > nelem:
            cosutil.printWarning("Flat field extends beyond the data array;"
                                 " the flat will be truncated.")
            if offset1 < 0:
                trim = -offset1
                flat[:int(offset1+len_ff)] = self.data_ff[int(trim):]
            else:               # offset1 + len_ff > nelem
                trim = (offset1 + len_ff) - nelem
                flat[int(offset1):int(offset1+len_ff-trim)] = self.data_ff[:int(len_ff-trim)]
        else:
            flat[int(offset1):int(offset1+len_ff)] = self.data_ff
        self.data_ff = self.convolveFlat(flat, doppmag, doppzero, orbitper,
                                         expstart, expend)

        self.state_ff = "1-D, matching data"

    def convolveFlat(self, flat, doppmag, doppzero, orbitper,
                     expstart, expend):
        """Convolve the flat with the Doppler smoothing function.

        Based on timetag.convolveFlat.

        Parameters
        ----------
        flat: array_like
            Flat field, collapsed to 1-D, float.

        doppmag: float
            Magnitude of Doppler shift in pixels.

        doppzero: float
            Time (MJD) when the orbital Doppler shift is zero and
            increasing.

        orbitper: float
            Orbital period (seconds) of HST.

        expstart: float
            Exposure start time (MJD).

        expend: float
            Exposure end time (MJD).

        Returns
        -------
            The flat field convolved with the Doppler shift throughout
            the exposure (i.e. from expstart to expend).  If the magnitude
            of the Doppler shift was zero, or expend is not later than
            expstart, the input flat will be returned.
        """

        if doppmag <= 0.:
            return flat

        # Round doppmag up to the next integer; mag is a zero-point offset.
        mag = int(math.ceil(doppmag + 1.))
        lendopp = 2 * mag + 1

        # dopp will be the normalized Doppler smoothing function.
        dopp = np.zeros(lendopp, dtype=np.float64)

        # This spans the exposure, may be greater than actual exposure time.
        exptime = (expend - expstart) * SEC_PER_DAY
        rnd_exptime = int(round(exptime))
        if rnd_exptime <= 0:
            return flat

        # t is the time in seconds since doppzero, in one-second increments.
        t = np.arange(rnd_exptime, dtype=np.float64) + \
                   (expstart - doppzero) * SEC_PER_DAY

        # shift is in pixels (wavelengths increase toward larger pixel number).
        shift = -doppmag * np.sin(2. * np.pi * t / orbitper)

        # Construct the Doppler smoothing function.
        i_npts = rnd_exptime
        i_npts = max(i_npts, 1)
        increment = 1. / float(i_npts)
        for i in range(i_npts):                     # one-second increments
            ishift = int(round(shift[i])) + mag
            dopp[ishift] += increment

        # Do the convolution (in-place).
        nelem = len(flat)
        conv_flat = np.zeros(nelem, dtype=np.float64)
        conv_flat[0:mag] = 1.
        conv_flat[nelem-mag:] = 1.
        for k in range(lendopp):
            low = k
            high = k + nelem - 2 * mag
            conv_flat[mag:nelem-mag] += (dopp[lendopp-1-k] * flat[low:high])

        return conv_flat

class OutputSpectrum(object):
    """An output spectrum.

    The interpolation and averaging for one row are done by invoking this.
    Data for the current output row are computed and assigned to the data
    block in ofd.

    Parameters
    ----------
    ofd: ``astropy.io.fits.hdu.hdulist.HDUList`` object
        For the output file

    inspec: list of Spectrum objects
        For the input tables

    keywords: dictionary
        keywords and values from input headers

    segment: str
        Segment or stripe name for current row

    output_wl_range: float
        Wavelength at first pixel

    output_dispersion: float
        Angstroms per pixel to use for output

    Attributes
    ----------
    ofd
    inspec
    keywords
    segment
    output_wl_range
    output_dispersion
    """

    def __init__(self, ofd, inspec, keywords, segment,
                  output_wl_range, output_dispersion):
        """Constructor."""

        self.ofd = ofd
        self.inspec = inspec
        self.keywords = keywords
        self.segment = segment

        data = self.ofd[1].data

        foundit = False
        for row in range(len(data)):
            if data.field("segment")[row] == self.segment:
                foundit = True
                break
        assert foundit == True

        nelem = data.field("nelem")[row]

        # Allocate space for the sum of weights.
        sumweight = np.zeros(nelem, dtype=np.float64)

        # Assign wavelengths for the current row.
        data.field("wavelength")[row,:] = output_wl_range[0] + \
                output_dispersion * np.arange(nelem, dtype=np.float64)

        for sp in self.inspec:
            if self.segment == sp.segment:
                self.accumulateSums(sp, data[row], sumweight)

        self.normalizeSums(data[row], sumweight)

    def normalizeSums(self, data, sumweight):
        """Divide the sums by the sum of the weights.

        Parameters
        ----------
        data: fits record array
            The current row of the output file

        sumweight: float
            Sum of weights
        """

        verylowweight = np.where(data.field("dq_wgt") < 0.1)
        sumweight = np.where(sumweight == 0., 1., sumweight)
        nonzeroweight = np.where(data.field("dq_wgt") != 0.)
        zeroweight = np.where(data.field("dq_wgt") == 0.)
        data.field("flux")[:] /= sumweight
        data.field("gross")[:] /= sumweight
        data.field("background")[:] /= sumweight
        variance = data.field("variance_flat") + data.field("variance_counts") + data.field("variance_bkg")
        variance = np.where(variance < 0.5, 0.5, variance)
        error_frequentist_lower, error_frequentist_upper = cosutil.errFrequentist(variance)
        error_frequentist_lower[zeroweight] = 0.0
        error_frequentist_upper[zeroweight] = 0.0
        bad = np.where(data.field("net") == 0.0)
        good = np.where(data.field("net") != 0.0)
        nelements = len(data.field("flux"))
        conversion = np.zeros(nelements)
        conversion[good] = data.field("flux")[good] / data.field("net")[good]
#       Clean out NaNs from where flux and net are zero
        wavelength = data.field("wavelength")
        interpolated_values = np.interp(wavelength[bad], wavelength[good], conversion[good])
        conversion[bad] = interpolated_values
        data.setfield("error_lower", error_frequentist_lower * conversion)
        data.setfield("error", error_frequentist_upper * conversion)
        data.field("net")[:] /= sumweight
        data.field("flux")[:][verylowweight] = 0.
        data.field("gross")[:][verylowweight] = 0.
        data.field("gcounts")[:][verylowweight] = 0.
        data.field("background")[:][verylowweight] = 0.
        data.field("variance_flat")[:][verylowweight] = 0.
        data.field("variance_counts")[:][verylowweight] = 0.
        data.field("variance_bkg")[:][verylowweight] = 0.
        data.field("net")[:][verylowweight] = 0.
        data.field("error_lower")[:][verylowweight] = 0.
        data.field("error")[:][verylowweight] = 0.
        data.field("dq_wgt")[:][verylowweight] = 0.

    def accumulateSums(self, sp, data, sumweight):
        """Add input data to output, weighting by exposure time.

        The values in data and sumweight will be modified in-place.
        This version allows for fractional-pixel offset of the input arrays.

        Parameters
        ----------
        sp: Spectrum object
            Current input spectrum

        data: ``astropy.io.fits.fitsrec.FITS_record`` object
            The current row of the output file

        sumweight: float
            Sum of weights
        """

        input_nelem = sp.nelem
        input_wavelength = sp.wavelength
        output_wavelength = data.field("wavelength")

        # Find the pixel numbers (floating point) in the input array
        # corresponding to the pixels in the output array, matching by
        # wavelength.  That is, for integer k:
        #   input_wavelength[ipixel[k]] = output_wavelength[k]
        # where the expression on the left hand side implies interpolation
        # rather than just indexing, since ipixel[k] will not in general
        # be an integer.

        ipixel = pixelsFromWl(input_wavelength, output_wavelength)

        # The output array will typically be longer than any of the input
        # arrays, so we must find the minimum and maximum indices in the
        # output array that map (via wavelength) to points within the current
        # input array.
        flag = np.where(np.logical_and (ipixel >= 0.,
                                        ipixel <= input_nelem-1.))
        min_k = flag[0][0]
        max_k = flag[0][-1]

        # ix is the array of pixel numbers (integer values but floating point
        # data type) in the input array that are actually within the input
        # array.  p and q are weight arrays for linear interpolation.
        ix = np.floor(ipixel[min_k:max_k])
        q = ipixel[min_k:max_k] - ix
        p = 1. - q
        i = ix.astype(np.int32)
        flux = data.field("flux")
        error = data.field("error")
        gross = data.field("gross")
        variance_flat = data.field("variance_flat")
        variance_counts = data.field("variance_counts")
        variance_bkg = data.field("variance_bkg")
        gcounts = data.field("gcounts")
        net = data.field("net")
        background = data.field("background")
        dq = data.field("dq")
        dq_wgt = data.field("dq_wgt")
        first = (data.field("exptime") == 0.)           # used for DQ

        weight1 = sp.dq_wgt[i] * sp.exptime
        weight2 = sp.dq_wgt[i+1] * sp.exptime
        # Also weight by the flat field.
        if sp.data_ff is not None:
            weight1 *= sp.data_ff[i]
            weight2 *= sp.data_ff[i+1]

        data.setfield("exptime", data.field("exptime") + sp.exptime)

        sumweight[min_k:max_k] += (p * weight1 + q * weight2)

        flux[min_k:max_k] += (sp.flux[i]   * p * weight1 +
                              sp.flux[i+1] * q * weight2)
        gross[min_k:max_k] += (sp.gross[i]   * p * weight1 +
                               sp.gross[i+1] * q * weight2)
        net[min_k:max_k] += (sp.net[i]   * p * weight1 +
                             sp.net[i+1] * q * weight2)
        background[min_k:max_k] += (sp.background[i]   * p * weight1 +
                                    sp.background[i+1] * q * weight2)
        temp_dq = sp.dq[i] | sp.dq[i+1]
        if first:
            dq[min_k:max_k] = temp_dq.copy()
        else:
            dq[min_k:max_k] &= temp_dq
        dq_wgt[min_k:max_k] += (sp.dq_wgt[i] * p + sp.dq_wgt[i+1] * q)
        gcounts[min_k:max_k] += (sp.gcounts[i]   * p * sp.dq_wgt[i] +
                                 sp.gcounts[i+1] * q * sp.dq_wgt[i+1])
        variance_flat[min_k:max_k] += (sp.variance_flat[i] * p * sp.dq_wgt[i] +
                                 sp.variance_flat[i+1] * q * sp.dq_wgt[i+1])
        variance_counts[min_k:max_k] += (sp.variance_counts[i] * p * sp.dq_wgt[i] +
                                 sp.variance_counts[i+1] * q * sp.dq_wgt[i+1])
        variance_bkg[min_k:max_k] += (sp.variance_bkg[i] * p * sp.dq_wgt[i] +
                                 sp.variance_bkg[i+1] * q * sp.dq_wgt[i+1])



"""Tests for COS/FUV timetag."""

import pytest

import calcos
from helpers import BaseCOS


# TODO: Mark this as slow when there are faster tests added for CI tests
#       so that this only runs in nightly tests.
@pytest.mark.slow
class TestFUVTimetag(BaseCOS):
    detector = 'fuv'

    def test_fuv_timetag_1(self):
        """
        FUV COS regression test #1
        """
        files_to_download = ['lckg01070_asn.fits', 'lckg01czq_spt.fits',
                             'lckg01d4q_spt.fits', 'lckg01d9q_spt.fits',
                             'lckg01dcq_spt.fits']

        # Prepare input files.
        self.get_input_files(files_to_download)

        # Run CALCOS
        input_file = 'lckg01070_asn.fits'
        calcos.calcos(input_file)

        # Compare results.
        # The first outroot is the output from whole ASN,
        # the rest are individual members.
        outroots = ['lckg01070', 'lckg01czq', 'lckg01d4q', 'lckg01d9q',
                    'lckg01dcq']
        outputs = []
        for sfx in ('x1dsum', 'x1dsum1', 'x1dsum2', 'x1dsum3', 'x1dsum4'):
            fname = '{}_{}.fits'.format(outroots[0], sfx)
            comparison_name = 'ref_' + fname
            outputs.append((fname, comparison_name))
        for outroot in outroots[1:]:
            for sfx in ('corrtag_a', 'corrtag_b', 'counts_a', 'counts_b',
                        'flt_a', 'flt_b', 'lampflash', 'x1d'):
                fname = '{}_{}.fits'.format(outroot, sfx)
                comparison_name = 'ref_' + fname
                outputs.append((fname, comparison_name))
        self.compare_outputs(outputs, rtol=3e-7)
